import functools
import random
import warnings
from pathlib import Path
from typing import Optional, List, Union, Callable, Literal

import numpy as np
from numpy.typing import NDArray

from audiomentations.core.audio_loading_utils import load_sound_file
from audiomentations.core.transforms_interface import BaseWaveformTransform
from audiomentations.core.utils import (
    calculate_desired_noise_rms,
    calculate_rms,
    calculate_rms_without_silence,
    convert_decibels_to_amplitude_ratio,
    find_audio_files_in_paths,
)


class AddShortNoises(BaseWaveformTransform):
    """Mix in various (bursts of overlapping) sounds with random pauses between. Useful if your
    original sound is clean, and you want to simulate an environment where short noises sometimes
    occur.
    A file or folder of (noise) sound(s) to be mixed in must be specified.
    """

    def __init__(
        self,
        sounds_path: Union[List[Path], List[str], Path, str],
        min_snr_db: float = -6.0,
        max_snr_db: float = 18.0,
        min_time_between_sounds: float = 2.0,
        max_time_between_sounds: float = 8.0,
        noise_rms: Literal["absolute", "relative", "relative_to_whole_input"] = "relative_to_whole_input",
        min_absolute_noise_rms_db: float = -50.0,
        max_absolute_noise_rms_db: float = -20.0,
        add_all_noises_with_same_level: bool = False,
        include_silence_in_noise_rms_estimation: bool = True,
        burst_probability: float = 0.22,
        min_pause_factor_during_burst: float = 0.1,
        max_pause_factor_during_burst: float = 1.1,
        min_fade_in_time: float = 0.005,
        max_fade_in_time: float = 0.08,
        min_fade_out_time: float = 0.01,
        max_fade_out_time: float = 0.1,
        signal_gain_in_db_during_noise: float = None,
        signal_gain_db_during_noise: float = None,
        noise_transform: Optional[
            Callable[[NDArray[np.float32], int], NDArray[np.float32]]
        ] = None,
        p: float = 0.5,
        lru_cache_size: Optional[int] = 64,
    ):
        """
        :param sounds_path: A path or list of paths to audio file(s) and/or folder(s) with
            audio files. Can be str or Path instance(s). The audio files given here are
            supposed to be (short) noises.
        :param min_snr_db: Minimum signal-to-noise ratio in dB. A lower value means the added
            sounds/noises will be louder. This gets ignored if noise_rms is set to "absolute".
        :param max_snr_db: Maximum signal-to-noise ratio in dB. A lower value means the added
            sounds/noises will be louder. This gets ignored if noise_rms is set to "absolute".
        :param min_time_between_sounds: Minimum pause time (in seconds) between the added sounds/noises
        :param max_time_between_sounds: Maximum pause time (in seconds) between the added sounds/noises
        :param noise_rms: Choices: ["absolute", "relative", "relative_to_whole_input"].
            Defines how the noises will be added to the audio input.
            "relative": the RMS value of the added noise will be proportional to the RMS value of
                the input sound calculated only for the region where the noise is added.
            "absolute": the added noises will have an RMS independent of the RMS of the input audio
                file.
            "relative_to_whole_input": the RMS of the added noises will be
                proportional to the RMS of the whole input sound.
        :param min_absolute_noise_rms_db: Is only used if noise_rms is set to "absolute". It is
            the minimum RMS value in dB that the added noise can take. The lower the RMS is, the
            lower will the added sound be.
        :param max_absolute_noise_rms_db: Is only used if noise_rms is set to "absolute". It is
            the maximum RMS value in dB that the added noise can take. Note that this value
            can not exceed 0.
        :param add_all_noises_with_same_level: Whether to add all the short noises
            (within one audio snippet) with the same SNR. If `noise_rms` is set to `"absolute"`,
            the RMS is used instead of SNR. The target SNR (or RMS) will change every time the
            parameters of the transform are randomized.
        :param include_silence_in_noise_rms_estimation: A boolean. It chooses how the RMS of
            the noises to be added will be calculated. If this option is set to False, the silence
            in the noise files will be disregarded in the RMS calculation. It is useful for
            non-stationary noises where silent periods occur.
        :param burst_probability: For every noise (A) that gets added, there
            is a probability of adding an extra noise (B) that overlaps with noise A. This
            parameter controls that probability. `min_pause_factor_during_burst` and
            `max_pause_factor_during_burst` control the amount of overlap.
        :param min_pause_factor_during_burst: Min value of how far into the current sound (as
            fraction) the burst sound should start playing. The value must be greater than 0.
        :param max_pause_factor_during_burst: Max value of how far into the current sound (as
            fraction) the burst sound should start playing. The value must be greater than 0.
        :param min_fade_in_time: Min sound/noise fade in time in seconds. Use a value larger
            than 0 to avoid a "click" at the start of the sound/noise.
        :param max_fade_in_time: Max sound/noise fade in time in seconds. Use a value larger
            than 0 to avoid a "click" at the start of the sound/noise.
        :param min_fade_out_time: Min sound/noise fade out time in seconds. Use a value larger
            than 0 to avoid a "click" at the end of the sound/noise.
        :param max_fade_out_time: Max sound/noise fade out time in seconds. Use a value larger
            than 0 to avoid a "click" at the end of the sound/noise.
        :param signal_gain_in_db_during_noise: Deprecated. Use signal_gain_db_during_noise instead.
        :param signal_gain_db_during_noise: Gain applied to the signal during a short noise.
            When fading the signal to the custom gain, the same fade times are used as
            for the noise, so it's essentially cross-fading. The default value (0.0) means
            the signal will not be gained. If set to a very low value, e.g. -100.0, this
            feature could be used for completely replacing the signal with the noise.
            This could be relevant in some use cases, for example:
            * replace the signal with another signal of a similar class (e.g. replace some
                speech with a cough)
            * simulate an ECG off-lead condition (electrodes are temporarily disconnected)
        :param noise_transform: A callable waveform transform (or composition of transforms) that
            gets applied to noises before they get mixed in.
        :param p: The probability of applying this transform
        :param lru_cache_size: Maximum size of the LRU cache for storing noise files in memory
        """
        super().__init__(p)
        self.sounds_path = sounds_path
        self.sound_file_paths = find_audio_files_in_paths(self.sounds_path)
        self.sound_file_paths = [str(p) for p in self.sound_file_paths]
        assert len(self.sound_file_paths) > 0

        assert min_time_between_sounds <= max_time_between_sounds
        assert 0.0 < burst_probability <= 1.0
        if burst_probability == 1.0:
            assert (
                min_pause_factor_during_burst > 0.0
            )  # or else an infinite loop will occur
        assert 0.0 < min_pause_factor_during_burst <= 1.0
        assert max_pause_factor_during_burst > 0.0
        assert max_pause_factor_during_burst >= min_pause_factor_during_burst
        assert min_fade_in_time >= 0.0
        assert max_fade_in_time >= 0.0
        assert min_fade_in_time <= max_fade_in_time
        assert min_fade_out_time >= 0.0
        assert max_fade_out_time >= 0.0
        assert min_fade_out_time <= max_fade_out_time
        assert min_absolute_noise_rms_db <= max_absolute_noise_rms_db < 0
        assert type(include_silence_in_noise_rms_estimation) == bool

        assert noise_rms in ["relative", "absolute", "relative_to_whole_input"]

        if min_snr_db > max_snr_db:
            raise ValueError("min_snr_db must not be greater than max_snr_db")
        self.min_snr_db = min_snr_db
        self.max_snr_db = max_snr_db
        self.signal_gain_in_db_during_noise = signal_gain_in_db_during_noise

        if (
            signal_gain_db_during_noise is not None
            and signal_gain_in_db_during_noise is not None
        ):
            raise ValueError(
                "Passing both signal_gain_db_during_noise and"
                " signal_gain_in_db_during_noise is not supported. Use only"
                " signal_gain_db_during_noise."
            )
        elif signal_gain_db_during_noise is not None:
            self.signal_gain_db_during_noise = signal_gain_db_during_noise
        elif signal_gain_in_db_during_noise is not None:
            warnings.warn(
                (
                    "The signal_gain_in_db_during_noise parameter is deprecated. Use"
                    " signal_gain_db_during_noise instead."
                ),
                DeprecationWarning,
            )
            self.signal_gain_db_during_noise = signal_gain_in_db_during_noise
        else:
            self.signal_gain_db_during_noise = 0.0  # the default

        self.min_time_between_sounds = min_time_between_sounds
        self.max_time_between_sounds = max_time_between_sounds
        self.burst_probability = burst_probability
        self.min_pause_factor_during_burst = min_pause_factor_during_burst
        self.max_pause_factor_during_burst = max_pause_factor_during_burst
        self.min_fade_in_time = min_fade_in_time
        self.max_fade_in_time = max_fade_in_time
        self.min_fade_out_time = min_fade_out_time
        self.max_fade_out_time = max_fade_out_time
        self.noise_rms = noise_rms
        self.min_absolute_noise_rms_db = min_absolute_noise_rms_db
        self.max_absolute_noise_rms_db = max_absolute_noise_rms_db
        self.include_silence_in_noise_rms_estimation = (
            include_silence_in_noise_rms_estimation
        )
        self.add_all_noises_with_same_level = add_all_noises_with_same_level
        self.noise_transform = noise_transform
        self.lru_cache_size = lru_cache_size
        self._load_sound = functools.lru_cache(maxsize=lru_cache_size)(
            AddShortNoises.__load_sound
        )

    @staticmethod
    def __load_sound(file_path, sample_rate):
        return load_sound_file(file_path, sample_rate)

    def randomize_parameters(self, samples: NDArray[np.float32], sample_rate: int):
        super().randomize_parameters(samples, sample_rate)
        if self.parameters["should_apply"]:
            input_sound_duration = len(samples) / sample_rate

            current_time = 0
            global_offset = random.uniform(
                -self.max_time_between_sounds, self.max_time_between_sounds
            )
            current_time += global_offset
            sounds = []

            snr_db = random.uniform(self.min_snr_db, self.max_snr_db)
            rms_db = random.uniform(
                self.min_absolute_noise_rms_db, self.max_absolute_noise_rms_db
            )

            while current_time < input_sound_duration:
                sound_file_path = random.choice(self.sound_file_paths)
                sound, _ = self.__load_sound(sound_file_path, sample_rate)
                sound_duration = len(sound) / sample_rate

                # Ensure that the fade time is not longer than the duration of the sound
                fade_in_time = min(
                    sound_duration,
                    random.uniform(self.min_fade_in_time, self.max_fade_in_time),
                )
                fade_out_time = min(
                    sound_duration,
                    random.uniform(self.min_fade_out_time, self.max_fade_out_time),
                )

                if not self.add_all_noises_with_same_level:
                    snr_db = random.uniform(self.min_snr_db, self.max_snr_db)
                    rms_db = random.uniform(
                        self.min_absolute_noise_rms_db, self.max_absolute_noise_rms_db
                    )

                sounds.append(
                    {
                        "fade_in_time": fade_in_time,
                        "start": current_time,
                        "end": current_time + sound_duration,
                        "fade_out_time": fade_out_time,
                        "file_path": sound_file_path,
                        "snr_db": snr_db,
                        "rms_db": rms_db,
                    }
                )

                # burst mode - add overlapping sounds
                while (
                    random.random() < self.burst_probability
                    and current_time < input_sound_duration
                ):
                    pause_factor = random.uniform(
                        self.min_pause_factor_during_burst,
                        self.max_pause_factor_during_burst,
                    )
                    pause_time = pause_factor * sound_duration
                    current_time = sounds[-1]["start"] + pause_time

                    if current_time >= input_sound_duration:
                        break

                    sound_file_path = random.choice(self.sound_file_paths)
                    sound, _ = self.__load_sound(sound_file_path, sample_rate)
                    sound_duration = len(sound) / sample_rate

                    fade_in_time = min(
                        sound_duration,
                        random.uniform(self.min_fade_in_time, self.max_fade_in_time),
                    )
                    fade_out_time = min(
                        sound_duration,
                        random.uniform(self.min_fade_out_time, self.max_fade_out_time),
                    )

                    if not self.add_all_noises_with_same_level:
                        snr_db = random.uniform(self.min_snr_db, self.max_snr_db)
                        rms_db = random.uniform(
                            self.min_absolute_noise_rms_db,
                            self.max_absolute_noise_rms_db,
                        )

                    sounds.append(
                        {
                            "fade_in_time": fade_in_time,
                            "start": current_time,
                            "end": current_time + sound_duration,
                            "fade_out_time": fade_out_time,
                            "file_path": sound_file_path,
                            "snr_db": snr_db,
                            "rms_db": rms_db,
                        }
                    )

                # wait until the last sound is done
                current_time += sound_duration

                # then add a pause
                pause_duration = random.uniform(
                    self.min_time_between_sounds, self.max_time_between_sounds
                )
                current_time += pause_duration

            self.parameters["sounds"] = sounds

    def apply(self, samples: NDArray[np.float32], sample_rate: int) -> NDArray[np.float32]:
        num_samples = samples.shape[-1]
        noise_placeholder = np.zeros_like(samples)

        signal_mask = None
        gain_signal = self.signal_gain_db_during_noise != 0.0
        if gain_signal:
            signal_mask = np.zeros(shape=(num_samples,), dtype=np.float32)

        for sound_params in self.parameters["sounds"]:
            if sound_params["end"] < 0:
                # Skip a sound if it ended before the start of the input sound
                continue

            noise_samples, _ = self.__load_sound(sound_params["file_path"], sample_rate)

            if self.noise_transform:
                noise_samples = self.noise_transform(noise_samples, sample_rate)

            # Apply fade in and fade out
            noise_gain = np.ones_like(noise_samples)
            fade_in_time_in_samples = int(sound_params["fade_in_time"] * sample_rate)
            fade_in_mask = np.linspace(0.0, 1.0, num=fade_in_time_in_samples)
            fade_out_time_in_samples = int(sound_params["fade_out_time"] * sample_rate)
            fade_out_mask = np.linspace(1.0, 0.0, num=fade_out_time_in_samples)
            noise_gain[: fade_in_mask.shape[0]] = fade_in_mask
            noise_gain[-fade_out_mask.shape[0] :] = np.minimum(
                noise_gain[-fade_out_mask.shape[0] :], fade_out_mask
            )

            start_sample_index = int(sound_params["start"] * sample_rate)
            end_sample_index = start_sample_index + noise_samples.shape[-1]

            if start_sample_index < 0:
                # crop noise_samples: shave off a chunk in the beginning
                num_samples_to_shave_off = abs(start_sample_index)
                noise_samples = noise_samples[num_samples_to_shave_off:]
                noise_gain = noise_gain[num_samples_to_shave_off:]
                start_sample_index = 0

            if end_sample_index > num_samples:
                # crop noise_samples: shave off a chunk in the end
                num_samples_to_shave_off = end_sample_index - num_samples
                end_index = noise_samples.shape[-1] - num_samples_to_shave_off
                noise_samples = noise_samples[:end_index]
                noise_gain = noise_gain[:end_index]
                end_sample_index = num_samples

            if self.noise_rms == "relative_to_whole_input":
                clean_rms = calculate_rms(samples)
            else:
                clean_rms = calculate_rms(samples[start_sample_index:end_sample_index])

            if noise_samples.shape[-1] > 0:
                # Gain here describes just the gain from the fade in and fade out.
                noise_samples = noise_samples * noise_gain

                if self.include_silence_in_noise_rms_estimation:
                    noise_rms = calculate_rms(noise_samples)
                else:
                    noise_rms = calculate_rms_without_silence(noise_samples, sample_rate)

                if noise_rms > 0:
                    if self.noise_rms in ["relative", "relative_to_whole_input"]:
                        desired_noise_rms = calculate_desired_noise_rms(
                            clean_rms, sound_params["snr_db"]
                        )

                        # Adjust the noise to match the desired noise RMS
                        noise_samples = noise_samples * (desired_noise_rms / noise_rms)
                    elif self.noise_rms == "absolute":
                        desired_noise_rms_db = sound_params["rms_db"]
                        desired_noise_rms_amp = convert_decibels_to_amplitude_ratio(
                            desired_noise_rms_db
                        )
                        gain = desired_noise_rms_amp / noise_rms
                        noise_samples = noise_samples * gain

                    noise_placeholder[start_sample_index:end_sample_index] += noise_samples
                    if gain_signal:
                        signal_mask[start_sample_index:end_sample_index] = np.maximum(
                            signal_mask[start_sample_index:end_sample_index],
                            noise_gain,
                        )

        if gain_signal:
            # Gain the original signal before mixing in the noises
            signal_mask *= self.signal_gain_db_during_noise
            signal_mask = convert_decibels_to_amplitude_ratio(signal_mask)
            return samples * signal_mask + noise_placeholder
        else:
            # Return a mix of the input sound and the added sounds
            return samples + noise_placeholder

    def __getstate__(self):
        state = self.__dict__.copy()
        warnings.warn(
            "Warning: the LRU cache of AddShortNoises gets discarded when pickling it."
            " E.g. this means the cache will not be used when using AddShortNoises"
            " together with multiprocessing on Windows"
        )
        del state["_load_sound"]
        return state



import numpy as np
import pytest

from audiomentations import Lambda, Gain


def test_gain_lambda():
    samples_in = np.random.normal(0, 1, size=1024).astype(np.float32)
    augmenter = Lambda(
        transform=Gain(min_gain_db=50, max_gain_db=50, p=1.0), p=1.0
    )
    std_in = np.mean(np.abs(samples_in))
    samples_out = augmenter(samples=samples_in, sample_rate=16000)
    std_out = np.mean(np.abs(samples_out))
    assert samples_out.dtype == np.float32
    assert samples_out.shape == samples_in.shape
    assert std_out > 100 * std_in

def test_lambda_with_kwargs():
    samples_in = np.random.normal(0, 1, size=1024).astype(np.float32)
    augmenter = Lambda(
        transform=lambda samples, sample_rate, offset: samples + offset,
        p=1.0,
        offset=-0.2,
    )
    input_mean = np.mean(samples_in)
    samples_out = augmenter(samples=samples_in, sample_rate=16000)
    output_mean = np.mean(samples_out)
    assert samples_out.dtype == np.float32
    assert samples_out.shape == samples_in.shape
    assert output_mean == pytest.approx(input_mean - 0.2)



import os

import numpy as np

from audiomentations import ApplyImpulseResponse
from audiomentations.core.composition import Compose
from demo.demo import DEMO_DIR


def test_apply_impulse_response():
    sample_len = 1024
    samples_in = np.random.normal(0, 1, size=sample_len).astype(np.float32)
    sample_rate = 16000

    add_ir_transform = ApplyImpulseResponse(ir_path=os.path.join(DEMO_DIR, "ir"), p=1.0)

    # Check that misc_file.txt is not one of the IR file candidates, as it's not audio
    assert len(add_ir_transform.ir_files) == 3

    augmenter = Compose([add_ir_transform])

    assert len(samples_in) == sample_len
    samples_out = augmenter(samples=samples_in, sample_rate=sample_rate)

    # Check parameters
    assert augmenter.transforms[0].parameters["should_apply"]
    assert augmenter.transforms[0].parameters["ir_file_path"].endswith(".wav")

    assert samples_out.dtype == np.float32
    assert samples_out.shape == samples_in.shape


def test_apply_impulse_response_multi_channel():
    sample_len = 1024
    samples_in = np.random.normal(0, 1, size=sample_len).astype(np.float32)
    sample_rate = 16000

    samples_in = np.expand_dims(samples_in, axis=0)
    samples_in = np.tile(samples_in, (2, 1))

    add_ir_transform = ApplyImpulseResponse(ir_path=os.path.join(DEMO_DIR, "ir"), p=1.0)

    # Check that misc_file.txt is not one of the IR file candidates, as it's not audio
    assert len(add_ir_transform.ir_files) == 3

    augmenter = Compose([add_ir_transform])

    assert samples_in.shape[1] == sample_len
    samples_out = augmenter(samples=samples_in, sample_rate=sample_rate)

    # Check parameters
    assert augmenter.transforms[0].parameters["should_apply"]
    assert augmenter.transforms[0].parameters["ir_file_path"].endswith(".wav")

    assert samples_out.dtype == np.float32
    assert samples_out.shape == samples_in.shape


def test_include_tail():
    sample_len = 1024
    samples_in = np.random.normal(0, 1, size=sample_len).astype(np.float32)
    sample_rate = 16000

    add_ir_transform = ApplyImpulseResponse(
        ir_path=os.path.join(DEMO_DIR, "ir"), p=1.0, leave_length_unchanged=False
    )

    samples_out = add_ir_transform(samples=samples_in, sample_rate=sample_rate)

    assert samples_out.dtype == np.float32
    assert samples_out.shape[-1] > samples_in.shape[-1]



import random

import numpy as np
import pytest

from audiomentations import AdjustDuration


@pytest.mark.parametrize("mode", ["silence", "wrap", "reflect"])
@pytest.mark.parametrize("pad_section", ["start", "end"])
@pytest.mark.parametrize("sample_len", [3, 4, 5])
@pytest.mark.parametrize("ndim", [None, 1, 2])
def test_padding(mode, pad_section, sample_len, ndim):
    random.seed(546)
    samples = np.ones((ndim, 4) if ndim else 4, dtype=np.float32)
    sample_rate = 16000
    input_shape = samples.shape
    target_shape = list(input_shape)
    target_shape[-1] = sample_len
    target_shape = tuple(target_shape)
    augmenter = AdjustDuration(
        duration_samples=sample_len,
        padding_mode=mode,
        padding_position=pad_section,
        p=1.0,
    )
    samples = augmenter(samples=samples, sample_rate=sample_rate)

    assert samples.dtype == np.float32
    assert samples.shape == target_shape


@pytest.mark.parametrize("mode", ["silence", "wrap", "reflect"])
@pytest.mark.parametrize("pad_section", ["start", "end"])
@pytest.mark.parametrize("second", [0.4, 0.5, 0.6])
@pytest.mark.parametrize("ndim", [None, 1, 2])
def test_padding_second(mode, pad_section, second, ndim):
    random.seed(546)
    sample_rate = 80
    samples = np.ones((ndim, 40) if ndim else 40, dtype=np.float32)
    input_shape = samples.shape
    target_shape = list(input_shape)
    target_shape[-1] = int(second * sample_rate)
    target_shape = tuple(target_shape)
    augmenter = AdjustDuration(
        duration_seconds=second, padding_mode=mode, padding_position=pad_section, p=1.0
    )
    samples = augmenter(samples=samples, sample_rate=sample_rate)

    assert samples.dtype == np.float32
    assert samples.shape == target_shape


def test_adjust_duration_raises_with_both_samples_and_seconds():
    with pytest.raises(ValueError) as exc_info:
        _ = AdjustDuration(duration_samples=100, duration_seconds=1.0)
    assert (
        "You must specify either duration_samples or duration_seconds, but not both."
        in str(exc_info.value)
    )


def test_adjust_duration_raises_with_non_positive_seconds():
    with pytest.raises(ValueError) as exc_info:
        _ = AdjustDuration(duration_seconds=0.0)
    assert "duration_seconds must be a positive float" in str(exc_info.value)

    with pytest.raises(ValueError) as exc_info:
        _ = AdjustDuration(duration_seconds=-1.0)
    assert "duration_seconds must be a positive float" in str(exc_info.value)


def test_adjust_duration_raises_with_non_positive_samples():
    with pytest.raises(ValueError) as exc_info:
        _ = AdjustDuration(duration_samples=0)
    assert "duration_samples must be a positive int" in str(exc_info.value)

    with pytest.raises(ValueError) as exc_info:
        _ = AdjustDuration(duration_samples=-1)
    assert "duration_samples must be a positive int" in str(exc_info.value)



import numpy as np
import pytest
import scipy
import scipy.signal

from audiomentations import LowShelfFilter
from tests.utils import get_chirp_test, get_randn_test

DEBUG = False


@pytest.mark.parametrize("center_freq", [2000.0, 3900.0])
@pytest.mark.parametrize("gain_db", [-6.0, +6.0, 0.0])
@pytest.mark.parametrize("q_factor", [0.1, 1.0])
def test_one_single_input(center_freq, gain_db, q_factor):
    np.random.seed(1)

    sample_rate = 16000

    # Parameters for computing periodograms.
    # When examining lower frequencies we need to have
    # a high nfft number.

    nfft = 4096
    nperseg = 1024

    samples = get_chirp_test(sample_rate, 40)

    augment = LowShelfFilter(
        min_center_freq=center_freq,
        max_center_freq=center_freq,
        min_gain_db=gain_db,
        max_gain_db=gain_db,
        min_q=q_factor,
        max_q=q_factor,
        p=1.0,
    )

    processed_samples = augment(samples, sample_rate=sample_rate)

    assert processed_samples.shape == samples.shape
    assert processed_samples.dtype == np.float32

    # Compute periodograms
    wx, samples_pxx = scipy.signal.welch(
        samples,
        fs=sample_rate,
        nfft=nfft,
        nperseg=nperseg,
        scaling="spectrum",
        window="hann",
    )
    _, processed_samples_pxx = scipy.signal.welch(
        processed_samples,
        fs=sample_rate,
        nperseg=nperseg,
        nfft=nfft,
        scaling="spectrum",
        window="hann",
    )

    if DEBUG:
        import matplotlib.pyplot as plt

        plt.title(f"Filter type: High-shelf at {gain_db} with Q={q_factor}")
        plt.plot(wx, 10 * np.log10(np.abs(samples_pxx)))
        plt.plot(wx, 10 * np.log10(np.abs(processed_samples_pxx)), ":")
        plt.legend(
            [
                "Input signal",
                f"Peaking with center freq {augment.parameters['center_freq']:.2f}",
            ]
        )
        plt.xlabel("Frequency (Hz)")
        plt.ylabel("Magnitude (dB)")
        plt.show()

    frequencies_of_interest = np.array([0.0, center_freq, sample_rate / 2])
    expected_differences = np.array([-gain_db, -gain_db / 2, 0.0])

    for n, freq in enumerate(frequencies_of_interest):
        input_value_db = 10 * np.log10(
            samples_pxx[int(np.round(nfft / sample_rate * freq))]
        )

        output_value_db = 10 * np.log10(
            processed_samples_pxx[int(np.round(nfft / sample_rate * freq))]
        )

        assert np.isclose(
            input_value_db - output_value_db,
            expected_differences[n],
            atol=1.0,
        )

@pytest.mark.parametrize("center_freq", [10.0, 2000.0, 3900.0])
@pytest.mark.parametrize("gain_db", [0.0, -6.0, +6.0])
@pytest.mark.parametrize("q_factor", [0.1, 1.0])
@pytest.mark.parametrize("num_channels", [1, 2, 3])
def test_n_channel_input(center_freq, gain_db, q_factor, num_channels):
    sample_rate = 8000
    samples = get_randn_test(sample_rate, 10)

    # Convert to 2D N channels
    n_channels = np.tile(samples, (num_channels, 1))

    augment = LowShelfFilter(
        min_center_freq=center_freq,
        max_center_freq=center_freq,
        min_gain_db=gain_db,
        max_gain_db=gain_db,
        min_q=q_factor,
        max_q=q_factor,
        p=1.0,
    )

    processed_samples = augment(samples=samples, sample_rate=sample_rate)

    processed_n_channels = augment(samples=n_channels, sample_rate=sample_rate)

    assert processed_n_channels.shape[0] == num_channels
    assert processed_n_channels.shape == n_channels.shape
    assert processed_n_channels.dtype == np.float32

    # Check that the processed 2D channel version applies the same effect
    # as the passband version.
    for _, channel in enumerate(processed_n_channels):
        if DEBUG:
            import matplotlib.pyplot as plt

            plt.title(f"Filter type: Low-shelf at {gain_db} with Q={q_factor}")

            plt.plot(processed_samples)
            plt.plot(channel, "r--")

            plt.legend(["1D", "2D"])
            plt.show()
        assert np.allclose(channel, processed_samples)



import random
from typing import Optional, Dict, Literal

import numpy as np
import sys
from numpy.typing import NDArray
from scipy.signal import convolve

from audiomentations.core.transforms_interface import BaseWaveformTransform


class RoomSimulator(BaseWaveformTransform):
    """
    A ShoeBox Room Simulator. Simulates a cuboid of parametrized size and 
    average surface absorption coefficient. It also includes a source
    and microphones in parametrized locations.

    Use it when you need a large number of synthetic room impulse responses with specific configuration
    characteristics or simply to quickly add reverb for augmentation purposes.

    Some examples:

    > augment = RoomSimulator(use_ray_tracing=False)
    Quickly add reverberation without caring for accuracy.

    > augment = RoomSimulator(calculation_mode="rt60", min_target_rt60=0.06, max_target_rt60=0.06, min_size_x = ...)
    Augment with randomly selected room impulse responses that have an RT60 of 0.06.

    > augment = RoomSimulator(min_mic_distance=1.0, max_mic_distance=1.0)
    Augment with a RIR captured by all positions of the microphone on a sphere, centred around the source at 1m

    > augment = RoomSimulator(min_mic_distance=1.0, max_mic_distance=1.0, min_mic_elevation=0.0, max_mic_elevation=0.0)
    Augment with a RIR captured by all positions of the microphone on a circle, centred around the source at 1m    
    """

    supports_multichannel = True

    def __init__(
        self,
        min_size_x: float = 3.6,
        max_size_x: float = 5.6,
        min_size_y: float = 3.6,
        max_size_y: float = 3.9,
        min_size_z: float = 2.4,
        max_size_z: float = 3.0,
        min_absorption_value: float = 0.075,
        max_absorption_value: float = 0.4,
        min_target_rt60: float = 0.15,
        max_target_rt60: float = 0.8,
        min_source_x: float = 0.1,
        max_source_x: float = 3.5,
        min_source_y: float = 0.1,
        max_source_y: float = 2.7,
        min_source_z: float = 1.0,
        max_source_z: float = 2.1,
        min_mic_distance: float = 0.15,
        max_mic_distance: float = 0.35,
        min_mic_azimuth: float = -np.pi,
        max_mic_azimuth: float = np.pi,
        min_mic_elevation: float = -np.pi,  # TODO: Check if it should default to -np.pi / 2 instead
        max_mic_elevation: float = np.pi,  # TODO: Check if it should default to np.pi / 2 instead
        calculation_mode: Literal["rt60", "absorption"] = "absorption",
        use_ray_tracing: bool = True,
        max_order: int = 1,
        leave_length_unchanged: bool = False,
        padding: float = 0.1,
        p: float = 0.5,
        ray_tracing_options: Optional[Dict] = None,
    ):
        """

        :param min_size_x: Minimum width (x coordinate) of the room in meters
        :param max_size_x: Maximum width of the room in meters
        :param min_size_y: Minimum depth (y coordinate) of the room in meters
        :param max_size_y: Maximum depth of the room in meters
        :param min_size_z: Minimum height (z coordinate) of the room in meters
        :param max_size_z: Maximum height of the room in meters
        :param min_absorption_value: When `calculation_mode` is "absorption" it will set
            a given coefficient value for the surfaces of the room (walls, ceilings, and floor).
            This coefficient takes values between 0 (fully reflective surface) and 1 (fully absorbing surface).

            Example values (may differ!):
                Studio w acoustic panels > 0.40
                Office / Library         ~ 0.15
                Factory                  ~ 0.05
        :param max_absorption_value:

        :param min_target_rt60: When `calculation_mode` is "rt60", it tries to set the absorption value
            of the surfaces of the room to achieve a target RT60 (in seconds). Note that this parameter
            changes only the materials (absorption coefficients) of the surfaces, NOT the dimension of the rooms.

            Example values (may differ!):
                Recording studio:  0.3s
                Office          :  0.5s
                Concert hall    :  1.5s

        :param min_source_x: Minimum x location of the source (meters)
        :param max_source_x: Minimum x location of the source (meters)
        :param min_source_y:
        :param max_source_y:
        :param min_source_z:
        :param max_source_z:
        :param min_mic_distance: Minimum distance of the microphone from the source in meters
        :param max_mic_distance:
        :param min_mic_azimuth: Minimum azimuth (angle around z axis) of the microphone
            relative to the source, in radians.
        :param max_mic_azimuth:
        :param min_mic_elevation: Minimum elevation of the microphone relative to the source, in radians.
        :param max_mic_elevation: Maximum elevation of the microphone relative to the source, in radians.
        :param calculation_mode: When set to `absorption`, it will create the room with surfaces based on
            `min_absorption_value` and `max_absorption_value`. If set to `rt60`, it will try to assign surface
            materials that lead to a room impulse response with the target RT60 given by `min_target_rt60` and `max_target_rt60`
        :param use_ray_tracing: Whether to use ray_tracing or not (slower but much more accurate).
            Disable this if you prioritize speed over accuracy.
        :param max_order: Maximum order of reflections for the Image Source Model. E.g. a value of
            1 will only add first order reflections while a value of 4-12 will add a
            diffuse reverberation tail. *Warning* Setting this higher than 11-12 can significantly slow down the
            augmentation process when `calculation_mode="rt60"`. In that case, keep it around `3-4`.
        :param leave_length_unchanged: When set to True, the tail of the sound (e.g. reverb at
            the end) will be chopped off so that the length of the output is equal to the
            length of the input.
        :param padding: Minimum distance in meters between source or mic and the room walls, floor or ceiling.
        :param p: The probability of applying this transform
        :param ray_tracing_options: Options for the ray tracer. See `set_ray_tracing` here:
            https://github.com/LCAV/pyroomacoustics/blob/master/pyroomacoustics/room.py
        """
        super().__init__(p)

        assert calculation_mode in [
            "rt60",
            "absorption",
        ], "`calculation_mode` should either be `rt60` or `absorption`"

        self.calculation_mode = calculation_mode
        self.min_absorption_value = min_absorption_value
        self.max_absorption_value = max_absorption_value

        self.min_target_rt60 = min_target_rt60
        self.max_target_rt60 = max_target_rt60

        self.use_ray_tracing = use_ray_tracing
        self.max_order = max_order

        self.min_size_x = min_size_x
        self.min_size_y = min_size_y
        self.min_size_z = min_size_z

        self.max_size_x = max_size_x
        self.max_size_y = max_size_y
        self.max_size_z = max_size_z

        assert min_source_x <= max_source_x
        assert min_source_y <= max_source_y
        assert min_source_z <= max_source_z

        self.min_source_x = min_source_x
        self.max_source_x = max_source_x

        self.min_source_y = min_source_y
        self.max_source_y = max_source_y

        self.min_source_z = min_source_z
        self.max_source_z = max_source_z

        assert min_mic_distance <= max_mic_distance
        assert min_mic_azimuth <= max_mic_azimuth
        assert min_mic_elevation <= max_mic_elevation

        self.min_mic_distance = min_mic_distance
        self.max_mic_distance = max_mic_distance

        self.min_mic_azimuth = min_mic_azimuth
        self.max_mic_azimuth = max_mic_azimuth

        self.min_mic_elevation = min_mic_elevation
        self.max_mic_elevation = max_mic_elevation

        self.leave_length_unchanged = leave_length_unchanged

        self.padding = padding

        if ray_tracing_options is None:
            self.ray_tracing_options = {
                "receiver_radius": 0.5,
                "n_rays": 10000,
                "energy_thres": 1e-5,
            }
        else:
            self.ray_tracing_options = ray_tracing_options

    def randomize_parameters(self, samples: NDArray[np.float32], sample_rate: int):

        try:
            import pyroomacoustics as pra
        except ImportError:
            print(
                "Failed to import pyroomacoustics. Maybe it is not installed? "
                "To install the optional pyroomacoustics dependency of audiomentations,"
                " do `pip install audiomentations[extras]` or simply "
                " `pip install pyroomacoustics`",
                file=sys.stderr,
            )
            raise

        super().randomize_parameters(samples, sample_rate)
        if not self.parameters["should_apply"]:
            return

        self.parameters["size_x"] = random.uniform(self.min_size_x, self.max_size_x)
        self.parameters["size_y"] = random.uniform(self.min_size_y, self.max_size_y)
        self.parameters["size_z"] = random.uniform(self.min_size_z, self.max_size_z)

        room_dim = np.array(
            [
                self.parameters["size_x"],
                self.parameters["size_y"],
                self.parameters["size_z"],
            ]
        )

        self.parameters["max_order"] = self.max_order

        if self.calculation_mode == "rt60":
            target_rt60 = random.uniform(self.min_target_rt60, self.max_target_rt60)
            self.parameters["target_rt60"] = target_rt60

            # If we are in rt60 mode, estimate the absorption coefficient on a desired target
            # rt60 value.
            self.parameters["absorption_coefficient"], max_order = pra.inverse_sabine(
                self.parameters["target_rt60"], room_dim
            )
            
            # When `rt60` is specified, use max_order from sabine's formula
            self.parameters["max_order"] = min(max_order,self.max_order)
        else:
            self.parameters["absorption_coefficient"] = random.uniform(
                self.min_absorption_value, self.max_absorption_value
            )

        self.parameters["source_x"] = random.uniform(
            max(self.min_source_x, self.padding),
            min(self.max_source_x, self.parameters["size_x"] - self.padding),
        )
        self.parameters["source_y"] = random.uniform(
            max(self.min_source_y, self.padding),
            min(self.max_source_y, self.parameters["size_y"] - self.padding),
        )
        self.parameters["source_z"] = random.uniform(
            max(self.min_source_z, self.padding),
            min(self.max_source_z, self.parameters["size_z"] - self.padding),
        )

        self.parameters["mic_radius"] = random.uniform(
            self.min_mic_distance, self.max_mic_distance
        )
        self.parameters["mic_azimuth"] = random.uniform(
            self.min_mic_azimuth, self.max_mic_azimuth
        )
        self.parameters["mic_elevation"] = random.uniform(
            self.min_mic_elevation, self.max_mic_elevation
        )

        # Convert to cartesian coordinates according to ADM
        mic_x = self.parameters["source_x"] - self.parameters["mic_radius"] * np.cos(
            self.parameters["mic_elevation"]
        ) * np.sin(self.parameters["mic_azimuth"])
        mic_y = self.parameters["source_y"] + self.parameters["mic_radius"] * np.cos(
            self.parameters["mic_elevation"]
        ) * np.cos(self.parameters["mic_azimuth"])
        mic_z = self.parameters["source_z"] + self.parameters["mic_radius"] * np.sin(
            self.parameters["mic_elevation"]
        )

        # Clamp between 0 and room dimensions
        self.parameters["mic_x"] = max(
            self.padding, min(self.parameters["size_x"] - self.padding, mic_x)
        )
        self.parameters["mic_y"] = max(
            self.padding, min(self.parameters["size_y"] - self.padding, mic_y)
        )
        self.parameters["mic_z"] = max(
            self.padding, min(self.parameters["size_z"] - self.padding, mic_z)
        )

        # Construct room
        self.room = pra.Room.from_corners(
            np.array(
                [
                    [0, 0],
                    [0, self.parameters["size_y"]],
                    [self.parameters["size_x"], self.parameters["size_y"]],
                    [self.parameters["size_x"], 0],
                ]
            ).T,
            fs=sample_rate,
            materials=pra.Material(self.parameters["absorption_coefficient"]),
            ray_tracing=self.use_ray_tracing,
            air_absorption=True,
            max_order=self.parameters["max_order"],
        )

        if self.use_ray_tracing:
            # TODO: Somehow make those parameters
            self.room.set_ray_tracing(**self.ray_tracing_options)

        self.room.extrude(
            height=self.parameters["size_z"],
            materials=pra.Material(self.parameters["absorption_coefficient"]),
        )

        # Add the point source
        self.room.add_source(
            np.array(
                [
                    self.parameters["source_x"],
                    self.parameters["source_y"],
                    self.parameters["source_z"],
                ]
            ),
            signal=samples,
        )

        # Add the microphone
        self.room.add_microphone_array(
            pra.MicrophoneArray(
                np.array(
                    [
                        [
                            self.parameters["mic_x"],
                            self.parameters["mic_y"],
                            self.parameters["mic_z"],
                        ]
                    ]
                ).T,
                self.room.fs,
            )
        )
        # Do the simulation
        self.room.compute_rir()

    def apply(self, samples: NDArray[np.float32], sample_rate: int) -> NDArray[np.float32]:
        assert samples.dtype == np.float32

        rir = self.room.rir[0][0]

        # This is the same as ApplyImpulseResponse transform
        if samples.ndim > 1:
            signal_ir = []
            for i in range(samples.shape[0]):
                channel_conv = convolve(samples[i], rir)
                signal_ir.append(channel_conv)
            signal_ir = np.array(signal_ir, dtype=samples.dtype)
        else:
            signal_ir = convolve(samples, rir).astype(samples.dtype)

        if self.leave_length_unchanged:
            signal_ir = signal_ir[..., : samples.shape[-1]]
        return signal_ir



import numpy as np
import pytest

from audiomentations import AddGaussianNoise


def test_gaussian_noise():
    samples = np.zeros((20,), dtype=np.float32)
    sample_rate = 16000
    augmenter = AddGaussianNoise(p=1.0)
    samples = augmenter(samples=samples, sample_rate=sample_rate)

    assert samples.dtype == np.float32
    assert not (float(np.sum(np.abs(samples))) == pytest.approx(0.0))


def test_gaussian_noise_stereo():
    samples = np.zeros((2, 2000), dtype=np.float32)
    sample_rate = 16000
    augmenter = AddGaussianNoise(p=1.0)
    samples = augmenter(samples=samples, sample_rate=sample_rate)

    assert samples.dtype == np.float32
    assert not (float(np.sum(np.abs(samples))) == pytest.approx(0.0))



import random
import sys

import numpy as np
from numpy.typing import NDArray

from audiomentations.core.transforms_interface import BaseWaveformTransform


class LoudnessNormalization(BaseWaveformTransform):
    """
    Apply a constant amount of gain to match a specific loudness (in LUFS). This is an
    implementation of ITU-R BS.1770-4.

    For an explanation on LUFS, see https://en.wikipedia.org/wiki/LUFS

    See also the following web pages for more info on audio loudness normalization:
        https://github.com/csteinmetz1/pyloudnorm
        https://en.wikipedia.org/wiki/Audio_normalization

    Warning: This transform can return samples outside the [-1, 1] range, which may lead to
    clipping or wrap distortion, depending on what you do with the audio in a later stage.
    See also https://en.wikipedia.org/wiki/Clipping_(audio)#Digital_clipping
    """

    supports_multichannel = True

    def __init__(
        self,
        min_lufs: float = -31.0,
        max_lufs: float = -13.0,
        p: float = 0.5,
    ):
        super().__init__(p)

        if min_lufs > max_lufs:
            raise ValueError("min_lufs must not be greater than max_lufs")

        self.min_lufs = min_lufs
        self.max_lufs = max_lufs

    def randomize_parameters(self, samples: NDArray[np.float32], sample_rate: int):
        try:
            import pyloudnorm
        except ImportError:
            print(
                (
                    "Failed to import pyloudnorm. Maybe it is not installed? "
                    "To install the optional pyloudnorm dependency of audiomentations,"
                    " do `pip install audiomentations[extras]` or simply "
                    " `pip install pyloudnorm`"
                ),
                file=sys.stderr,
            )
            raise

        super().randomize_parameters(samples, sample_rate)
        if self.parameters["should_apply"]:
            meter = pyloudnorm.Meter(sample_rate)  # create BS.1770 meter
            # Transpose because pyloudnorm expects shape like (smp, chn), not (chn, smp)
            self.parameters["loudness"] = float(
                meter.integrated_loudness(samples.transpose())
            )
            self.parameters["lufs"] = float(
                random.uniform(self.min_lufs, self.max_lufs)
            )

    def apply(
        self, samples: NDArray[np.float32], sample_rate: int
    ) -> NDArray[np.float32]:
        try:
            import pyloudnorm
        except ImportError:
            print(
                (
                    "Failed to import pyloudnorm. Maybe it is not installed? "
                    "To install the optional pyloudnorm dependency of audiomentations,"
                    " do `pip install audiomentations[extras]` or simply "
                    " `pip install pyloudnorm`"
                ),
                file=sys.stderr,
            )
            raise

        # Guard against digital silence
        if self.parameters["loudness"] > float("-inf"):
            # Normalize loudness
            delta_loudness = self.parameters["lufs"] - self.parameters["loudness"]
            gain = np.power(10.0, delta_loudness / 20.0, dtype=np.float32)
            return gain * samples

        return samples



import random
import numpy as np
from typing import Literal
from numpy.typing import NDArray

from audiomentations.core.transforms_interface import BaseWaveformTransform
from audiomentations.core.utils import get_crossfade_mask_pair

# 0.00025 s corresponds to 2 samples at 8 kHz
DURATION_EPSILON = 0.00025


class TimeMask(BaseWaveformTransform):
    """
    Make a chosen part of the audio silent (time-masking augmentation).

    The silent part can optionally be faded in/out to avoid hard transients.

    Inspired by *SpecAugment: A Simple Data Augmentation Method for Automatic
    Speech Recognition* (https://arxiv.org/pdf/1904.08779.pdf)
    """

    supports_multichannel = True

    def __init__(
        self,
        min_band_part: float = 0.01,
        max_band_part: float = 0.2,
        fade_duration: float = 0.005,
        mask_location: Literal["start", "end", "random"] = "random",
        p: float = 0.5,
        **kwargs,
    ):
        """
        :param min_band_part: Minimum length of the silent part as a fraction of the
            total sound length. Must be between 0.0 and 1.0
        :param max_band_part: Maximum length of the silent part as a fraction of the
            total sound length. Must be between 0.0 and 1.0
        :param fade_duration: Duration (in seconds) of the fade-in and fade-out applied
            at the edges of the silent region to smooth transitions and avoid abrupt
            changes, which can otherwise produce impulses or clicks in the audio.
            If you need hard edges or clicks, set this to `0.0` to disable fading.
            Positive values must be at least 0.00025 seconds.
        :param mask_location: Where to place the silent region.
            * "start": silence begins at index 0
            * "end": silence ends at the last sample
            * "random" (default): silence starts at a random position
        :param p: The probability of applying this transform
        """
        if "fade" in kwargs:
            raise TypeError(
                "The 'fade' parameter was removed in v0.41.0 and is no longer supported."
                " Please use the 'fade_duration' parameter (float, seconds) instead."
                " To disable fading (equivalent to fade=False), set fade_duration=0.0."
                " To enable fading (equivalent to fade=True), set fade_duration to a"
                " positive value (e.g., the default 0.005 seconds)."
            )

        super().__init__(p)

        if not (0.0 <= min_band_part <= 1.0):
            raise ValueError("min_band_part must be between 0.0 and 1.0")
        if not (0.0 <= max_band_part <= 1.0):
            raise ValueError("max_band_part must be between 0.0 and 1.0")
        if min_band_part > max_band_part:
            raise ValueError("min_band_part must not be greater than max_band_part")

        if fade_duration < 0:
            raise ValueError("fade_duration must be non-negative")
        if 0 < fade_duration < DURATION_EPSILON:
            raise ValueError(
                f"When fade_duration is positive it must be >= {DURATION_EPSILON}"
            )

        if mask_location not in {"start", "end", "random"}:
            raise ValueError('mask_location must be "start", "end" or "random"')

        self.min_band_part = min_band_part
        self.max_band_part = max_band_part
        self.fade_duration = fade_duration
        self.mask_location = mask_location

    def randomize_parameters(self, samples: NDArray[np.float32], sample_rate: int):
        super().randomize_parameters(samples, sample_rate)
        if not self.parameters["should_apply"]:
            return

        num_samples = samples.shape[-1]

        # Mask length
        t = random.randint(
            int(num_samples * self.min_band_part),
            int(num_samples * self.max_band_part),
        )

        # Start index based on mask_location
        loc = self.mask_location
        if loc == "start":
            t0 = 0
        elif loc == "end":
            t0 = num_samples - t
        else:  # "random"
            t0 = random.randint(0, num_samples - t)

        self.parameters.update({"t": t, "t0": t0})

    def apply(
        self, samples: NDArray[np.float32], sample_rate: int
    ) -> NDArray[np.float32]:
        new_samples = samples.copy()
        t: int = self.parameters["t"]
        t0: int = self.parameters["t0"]

        fade_len = 0
        if self.fade_duration > 0.0:
            fade_len = int(round(sample_rate * self.fade_duration))
            fade_len = min(fade_len, t // 2)

        if fade_len >= 2:
            fade_in, fade_out = get_crossfade_mask_pair(fade_len, equal_energy=False)

            left = slice(t0, t0 + fade_len)
            mid = slice(t0 + fade_len, t0 + t - fade_len)
            right = slice(t0 + t - fade_len, t0 + t)

            new_samples[..., left] *= fade_out
            if mid.start < mid.stop:
                new_samples[..., mid] = 0.0
            new_samples[..., right] *= fade_in
        else:
            new_samples[..., t0 : t0 + t] = 0.0

        return new_samples



import os

import numpy as np
from numpy.testing import assert_array_equal

from audiomentations import (
    ClippingDistortion,
    AddBackgroundNoise,
    TimeMask,
    Shift,
    Compose,
)
from demo.demo import DEMO_DIR


def test_freeze_and_unfreeze_parameters():
    samples = np.zeros((20,), dtype=np.float32)
    sample_rate = 44100
    augmenter = Compose(
        [
            AddBackgroundNoise(
                sounds_path=os.path.join(DEMO_DIR, "background_noises"),
                min_snr_db=15,
                max_snr_db=35,
                p=1.0,
            ),
            ClippingDistortion(p=0.5),
        ]
    )
    perturbed_samples1 = augmenter(samples=samples, sample_rate=sample_rate)
    augmenter.freeze_parameters()
    for transform in augmenter.transforms:
        assert transform.are_parameters_frozen
    perturbed_samples2 = augmenter(samples=samples, sample_rate=sample_rate)

    assert_array_equal(perturbed_samples1, perturbed_samples2)

    augmenter.unfreeze_parameters()
    for transform in augmenter.transforms:
        assert not transform.are_parameters_frozen

def test_randomize_parameters_and_apply():
    samples = 1.0 / np.arange(1, 21, dtype=np.float32)
    sample_rate = 44100

    augmenter = Compose(
        [
            AddBackgroundNoise(
                sounds_path=os.path.join(DEMO_DIR, "background_noises"),
                min_snr_db=15,
                max_snr_db=35,
                p=1.0,
            ),
            ClippingDistortion(p=0.5),
            TimeMask(min_band_part=0.2, max_band_part=0.5, p=0.5),
            Shift(min_shift=0.5, max_shift=0.5, p=0.5),
        ]
    )
    augmenter.freeze_parameters()
    augmenter.randomize_parameters(samples=samples, sample_rate=sample_rate)

    parameters = [transform.parameters for transform in augmenter.transforms]

    perturbed_samples1 = augmenter(samples=samples, sample_rate=sample_rate)
    perturbed_samples2 = augmenter(samples=samples, sample_rate=sample_rate)

    assert_array_equal(perturbed_samples1, perturbed_samples2)

    augmenter.unfreeze_parameters()

    for transform_parameters, transform in zip(parameters, augmenter.transforms):
        assert transform_parameters == transform.parameters
        assert not transform.are_parameters_frozen



import os
import random
import warnings

import numpy as np
import pytest

from audiomentations import AddBackgroundNoise, Reverse
from demo.demo import DEMO_DIR


def test_add_background_noise():
    samples = np.sin(np.linspace(0, 440 * 2 * np.pi, 22500)).astype(np.float32)
    sample_rate = 44100
    augmenter = AddBackgroundNoise(
        sounds_path=os.path.join(DEMO_DIR, "background_noises"),
        min_snr_db=15.0,
        max_snr_db=35.0,
        p=1.0,
    )
    samples_out = augmenter(samples=samples, sample_rate=sample_rate)
    assert not np.allclose(samples, samples_out)
    assert samples_out.dtype == np.float32


def test_add_background_noise_when_noise_sound_is_too_short():
    sample_rate = 44100
    samples = np.sin(np.linspace(0, 440 * 2 * np.pi, 14 * sample_rate)).astype(
        np.float32
    )
    augmenter = AddBackgroundNoise(
        sounds_path=os.path.join(DEMO_DIR, "background_noises"),
        min_snr_db=15,
        max_snr_db=35,
        p=1.0,
    )
    samples_out = augmenter(samples=samples, sample_rate=sample_rate)
    assert not np.allclose(samples, samples_out)
    assert samples_out.dtype == np.float32


def test_try_add_almost_silent_file():
    samples = np.sin(np.linspace(0, 440 * 2 * np.pi, 30000)).astype(np.float32)
    sample_rate = 48000
    augmenter = AddBackgroundNoise(
        sounds_path=os.path.join(DEMO_DIR, "almost_silent"),
        min_snr_db=15,
        max_snr_db=35,
        p=1.0,
    )
    samples_out = augmenter(samples=samples, sample_rate=sample_rate)
    assert not np.allclose(samples, samples_out)
    assert samples_out.dtype == np.float32


def test_try_add_digital_silence():
    samples = np.sin(np.linspace(0, 440 * 2 * np.pi, 40000)).astype(np.float32)
    sample_rate = 48000
    augmenter = AddBackgroundNoise(
        sounds_path=os.path.join(DEMO_DIR, "digital_silence"),
        min_snr_db=15,
        max_snr_db=35,
        p=1.0,
    )

    with warnings.catch_warnings(record=True) as w:
        # Cause all warnings to always be triggered.
        warnings.simplefilter("always")
        samples_out = augmenter(samples=samples, sample_rate=sample_rate)

        assert "is too silent to be added as noise" in str(w[-1].message)

    assert np.allclose(samples, samples_out)
    assert samples_out.dtype == np.float32


def test_absolute_option():
    samples = np.sin(np.linspace(0, 440 * 2 * np.pi, 22500)).astype(np.float32)
    sample_rate = 44100
    augmenter = AddBackgroundNoise(
        sounds_path=os.path.join(DEMO_DIR, "background_noises"),
        noise_rms="absolute",
        p=1.0,
    )
    samples_out = augmenter(samples=samples, sample_rate=sample_rate)
    assert not np.allclose(samples, samples_out)


def test_noise_transform():
    np.random.seed(3650)
    random.seed(3650)
    samples = np.sin(np.linspace(0, 440 * 2 * np.pi, 22500)).astype(np.float32)
    sample_rate = 44100
    augmenter = AddBackgroundNoise(
        sounds_path=os.path.join(DEMO_DIR, "background_noises"),
        min_snr_db=3,
        max_snr_db=6,
        p=1.0,
    )
    samples_out_without_transform = augmenter(samples=samples, sample_rate=sample_rate)
    augmenter.freeze_parameters()
    augmenter.noise_transform = Reverse()
    samples_out_with_transform = augmenter(samples=samples, sample_rate=sample_rate)

    assert not np.allclose(samples_out_without_transform, samples_out_with_transform)


def test_validation():
    with pytest.raises(ValueError):
        AddBackgroundNoise(
            sounds_path=os.path.join(DEMO_DIR, "background_noises"),
            min_snr_db=45.0,
            max_snr_db=35.0,
        )
    with pytest.raises(ValueError):
        AddBackgroundNoise(
            sounds_path=os.path.join(DEMO_DIR, "background_noises"),
            min_absolute_rms_db=-1.0,
            max_absolute_rms_db=-5.0,
        )
    with pytest.raises(ValueError):
        AddBackgroundNoise(
            sounds_path=os.path.join(DEMO_DIR, "background_noises"),
            min_absolute_rms_db=-1.0,
            max_absolute_rms_db=10.0,
        )



import os

import numpy as np
import pytest

from audiomentations.core.utils import (
    calculate_desired_noise_rms,
    convert_decibels_to_amplitude_ratio,
    find_audio_files_in_paths,
    calculate_rms,
    calculate_rms_without_silence,
    a_weighting_frequency_envelope,
    get_crossfade_length,
)
from demo.demo import DEMO_DIR


def test_calculate_desired_noise_rms():
    noise_rms = calculate_desired_noise_rms(clean_rms=0.5, snr=6)
    assert noise_rms == pytest.approx(0.2505936168136362)


def test_convert_decibels_to_amplitude_ratio():
    amplitude_ratio = convert_decibels_to_amplitude_ratio(decibels=-6)
    assert amplitude_ratio == pytest.approx(0.5011872336272722)

    amplitude_ratio = convert_decibels_to_amplitude_ratio(decibels=6)
    assert amplitude_ratio == pytest.approx(1.9952623149688795)


def test_find_audio_files_in_paths_uppercase_extension():
    file_paths = find_audio_files_in_paths(DEMO_DIR, traverse_subdirectories=False)
    found_it = False
    for file_path in file_paths:
        if file_path.name == "stereo_24bit.WAV":
            found_it = True
            break
    assert found_it


def test_find_single_audio_file():
    file_paths = find_audio_files_in_paths(
        os.path.join(DEMO_DIR, "bus.opus"), traverse_subdirectories=False
    )
    assert len(file_paths) == 1
    assert file_paths[0].name == "bus.opus"


def test_find_multiple_audio_files():
    file_paths = find_audio_files_in_paths(
        [os.path.join(DEMO_DIR, "bus.opus"), os.path.join(DEMO_DIR, "testing.m4a")]
    )
    assert len(file_paths) == 2
    assert file_paths[0].name == "bus.opus"
    assert file_paths[1].name == "testing.m4a"


def test_find_audio_files_multiple_dirs():
    file_paths = find_audio_files_in_paths(
        [os.path.join(DEMO_DIR, "short_noises"), os.path.join(DEMO_DIR, "ir")]
    )
    assert len(file_paths) == 8
    assert file_paths[0].name == "130921_iPhone_rub_channel0_chunk83_aug2.wav"


def test_calculate_rms_stereo():
    np.random.seed(42)
    sample_len = 1024
    samples_in = np.random.uniform(low=-0.5, high=0.5, size=(2, sample_len)).astype(
        np.float32
    )
    rms = calculate_rms(samples_in)
    assert rms == pytest.approx(0.287, abs=0.01)


def test_calculate_rms_without_silence():
    sample_rate = 48000
    samples_in = np.zeros(int(2.0022 * sample_rate))
    samples_in[0:sample_rate] = 0.4 * np.ones(sample_rate)
    rms_before = calculate_rms(samples_in)
    rms_after = calculate_rms_without_silence(samples_in, sample_rate)
    assert rms_after > rms_before
    assert rms_after == pytest.approx(0.4)

    # Check that the function works if the input is shorter than a window (25 ms)
    rms_short = calculate_rms_without_silence(
        samples_in[0 : int(0.015 * sample_rate)], sample_rate
    )
    assert rms_short == pytest.approx(0.4)


@pytest.mark.parametrize(
    "sample_rate, n_fft",
    [
        (16000, 1024),
        (48000, 4096),
    ],
)
def test_a_weighting_curve(sample_rate, n_fft):
    freqs = np.fft.rfftfreq(n_fft, 1 / sample_rate)
    weighting = (
        (12194**2 * freqs**4)
        / (
            (freqs**2 + 20.6**2)
            * np.sqrt((freqs**2 + 107.7**2) * (freqs**2 + 737.9**2))
            * (freqs**2 + 12194**2)
        )
    ) + 2.00

    assert a_weighting_frequency_envelope(n_fft, sample_rate) == pytest.approx(
        weighting, abs=0.01
    )


def test_get_crossfade_duration_too_short_duration():
    with pytest.warns(UserWarning, match="crossfade_duration is too small"):
        result = get_crossfade_length(sample_rate=44_100, crossfade_duration=0.00001)
    assert result == 2


def test_get_crossfade_duration_odd_length_case():
    result = get_crossfade_length(sample_rate=10, crossfade_duration=0.3)
    # The function should add 1 to 3, resulting in an even value of 4.
    assert result == 4



import random

import numpy as np
from numpy.typing import NDArray

from audiomentations.core.transforms_interface import BaseWaveformTransform
from audiomentations.core.utils import convert_decibels_to_amplitude_ratio


class Gain(BaseWaveformTransform):
    """
    Multiply the audio by a random amplitude factor to reduce or increase the volume. This
    technique can help a model become somewhat invariant to the overall gain of the input audio.

    Warning: This transform can return samples outside the [-1, 1] range, which may lead to
    clipping or wrap distortion, depending on what you do with the audio in a later stage.
    See also https://en.wikipedia.org/wiki/Clipping_(audio)#Digital_clipping
    """

    supports_multichannel = True

    def __init__(
        self,
        min_gain_db: float = -12.0,
        max_gain_db: float = 12.0,
        p: float = 0.5,
    ):
        """
        :param min_gain_db: Minimum gain
        :param max_gain_db: Maximum gain
        :param p: The probability of applying this transform
        """
        super().__init__(p)

        if min_gain_db > max_gain_db:
            raise ValueError("min_gain_db cannot be greater than max_gain_db")
        self.min_gain_db = min_gain_db
        self.max_gain_db = max_gain_db

    def randomize_parameters(self, samples: NDArray[np.float32], sample_rate: int):
        super().randomize_parameters(samples, sample_rate)
        if self.parameters["should_apply"]:
            self.parameters["amplitude_ratio"] = convert_decibels_to_amplitude_ratio(
                random.uniform(self.min_gain_db, self.max_gain_db)
            )

    def apply(self, samples: NDArray[np.float32], sample_rate: int) -> NDArray[np.float32]:
        return samples * self.parameters["amplitude_ratio"]



from copy import deepcopy

import numpy as np
import pytest
from numpy.testing import assert_array_equal

from audiomentations import PitchShift


@pytest.mark.parametrize("method", ["librosa_phase_vocoder", "signalsmith_stretch"])
def test_1dim(method):
    samples = np.zeros((2048,), dtype=np.float32)
    sample_rate = 16000
    augmenter = PitchShift(min_semitones=-2, max_semitones=-1, method=method, p=1.0)
    samples_out = augmenter(samples=samples, sample_rate=sample_rate)

    assert samples_out.dtype == np.float32
    assert samples_out.shape == (2048,)


@pytest.mark.parametrize("method", ["librosa_phase_vocoder", "signalsmith_stretch"])
def test_multichannel(method):
    num_channels = 3
    samples = np.random.normal(0, 0.1, size=(num_channels, 5555)).astype(np.float32)
    sample_rate = 16000
    augmenter = PitchShift(min_semitones=1, max_semitones=2, method=method, p=1.0)
    samples_out = augmenter(samples=samples, sample_rate=sample_rate)

    assert samples_out.dtype == np.float32
    assert samples_out.shape == samples.shape
    for i in range(num_channels):
        assert not np.allclose(samples[i], samples_out[i])


def test_stereo_non_contiguous_ndarray():
    num_channels = 2
    samples = np.random.normal(0, 0.1, size=(5555, num_channels)).astype(np.float32)
    sample_rate = 16000
    augmenter = PitchShift(
        min_semitones=1, max_semitones=2, method="signalsmith_stretch", p=1.0
    )
    samples_out = augmenter(samples=samples.T, sample_rate=sample_rate)

    assert samples_out.dtype == np.float32
    assert samples_out.shape == samples.T.shape


def test_freeze_parameters():
    """
    Test that the transform can freeze its parameters, e.g. to apply the effect with the
    same parameters to multiple sounds.
    """
    samples = np.sin(np.linspace(0, 440 * 2 * np.pi, 8000)).astype(np.float32)
    sample_rate = 16000
    augmenter = PitchShift(min_semitones=1, max_semitones=12, p=1.0)

    first_samples = augmenter(samples=samples, sample_rate=sample_rate)
    first_parameters = deepcopy(augmenter.parameters)

    augmenter.min_semitones = -12
    augmenter.max_semitones = -1
    augmenter.are_parameters_frozen = True
    second_samples = augmenter(samples=samples, sample_rate=sample_rate)

    assert first_parameters == augmenter.parameters
    assert_array_equal(first_samples, second_samples)


def test_validate_parameters():
    with pytest.raises(ValueError):
        PitchShift(min_semitones=-40)
    with pytest.raises(ValueError):
        PitchShift(max_semitones=40)
    with pytest.raises(ValueError):
        PitchShift(min_semitones=5, max_semitones=2)
    with pytest.raises(ValueError):
        PitchShift(method="invalidvalue")



import numpy as np
import pytest

from audiomentations import Aliasing


def test_single_channel():
    samples = np.random.normal(0, 0.1, size=(2048,)).astype(np.float32)
    sample_rate = 16000
    augmenter = Aliasing(min_sample_rate=8000, max_sample_rate=32000, p=1.0)

    distorted_samples = augmenter(samples=samples, sample_rate=sample_rate)

    assert samples.dtype == distorted_samples.dtype
    assert samples.shape == distorted_samples.shape
    assert not np.array_equal(samples, distorted_samples)


def test_multichannel():
    num_channels = 3
    samples = np.random.normal(0, 0.1, size=(num_channels, 2048)).astype(np.float32)
    sample_rate = 16000
    augmenter = Aliasing(min_sample_rate=8000, max_sample_rate=32000, p=1.0)

    distorted_samples = augmenter(samples=samples, sample_rate=sample_rate)

    assert samples.dtype == distorted_samples.dtype
    assert samples.shape == distorted_samples.shape
    assert not np.array_equal(samples, distorted_samples)


def test_param_range():
    with pytest.raises(ValueError):
        Aliasing(min_sample_rate=0, max_sample_rate=6000, p=1.0)
    with pytest.raises(ValueError):
        Aliasing(min_sample_rate=8000, max_sample_rate=6000, p=1.0)



import numpy as np
import pytest

from audiomentations import Clip


def test_single_channel():
    samples = np.array([0.5, 0.6, -0.2, 0.0], dtype=np.float32)
    sample_rate = 16000
    augmenter = Clip(a_min=-0.1, a_max=0.1, p=1.0)
    samples = augmenter(samples=samples, sample_rate=sample_rate)

    assert np.amin(samples) == pytest.approx(-0.1)
    assert np.amax(samples) == pytest.approx(0.1)
    assert samples.dtype == np.float32
    assert samples.shape[-1] == 4


def test_multichannel():
    samples = np.array(
        [[0.9, 0.5, -0.25, -0.125, 0.0], [0.95, 0.5, -0.25, -0.125, 0.0]],
        dtype=np.float32,
    )
    sample_rate = 16000
    augmenter = Clip(a_min=-0.1, a_max=0.1, p=1.0)
    samples = augmenter(samples=samples, sample_rate=sample_rate)

    assert np.amin(samples) == pytest.approx(-0.1)
    assert np.amax(samples) == pytest.approx(0.1)
    assert samples.dtype == np.float32



from audiomentations.augmentations.add_color_noise import generate_decaying_white_noise
from audiomentations.core.utils import calculate_rms
import pytest

import numpy as np

from audiomentations import AddColorNoise, NOISE_COLOR_DECAYS


def noise_kl_divergence_from_white_noise(noise):
    """Measures similarity between noise and white noise using KL divergence"""

    def kl_divergence(p, q):
        return sum(p[i] * np.log2(p[i] / q[i]) for i in range(len(p)))

    hist1, _ = np.histogram(noise, bins=min(100, int(0.01 * len(noise))))
    hist1 = hist1.astype(np.float32) + 1e-4
    hist1 /= hist1.sum()

    white_noise_of_same_size = np.random.normal(0, 1, len(noise))
    hist2, _ = np.histogram(
        white_noise_of_same_size, bins=min(100, int(0.01 * len(noise)))
    )
    hist2 = hist2.astype(np.float32) + 1e-4
    hist2 /= hist2.sum()

    return kl_divergence(hist1, hist2)


def calculate_decay_rate(noise: np.ndarray, n_fft=8192):
    """Calculates the decay rate per octave of a noise PSD"""

    pxx = np.square(np.abs(np.fft.rfft(noise, n=n_fft)))
    pxx = pxx / np.sum(pxx)
    f = np.arange(len(pxx)).astype(np.float32)
    f[0] = 1e-10

    x = np.log10(f)

    # Linear regression using polyfit
    m, _ = np.polyfit(x[1:], 10 * np.log10(pxx[1:]), 1)

    return m * np.log10(2)


def test_add_colored_noise_defaults():
    np.random.seed(42)
    samples_in = np.random.normal(0, 1, size=1024).astype(np.float32)
    augmenter = AddColorNoise(p=1.0)
    rms_before = np.sqrt(np.mean(samples_in**2))
    samples_out = augmenter(samples=samples_in, sample_rate=16000)
    rms_after = np.sqrt(np.mean(samples_out**2))
    assert samples_out.shape == samples_in.shape
    assert samples_out.dtype == np.float32
    assert not (float(rms_after) == pytest.approx(0.0))
    assert rms_after != rms_before


def test_add_colored_noise_defaults_stereo():
    np.random.seed(42)
    samples_in = np.random.normal(0, 1, size=(2, 1024)).astype(np.float32)
    augmenter = AddColorNoise(p=1.0)
    rms_before = np.sqrt(np.mean(samples_in**2))
    samples_out = augmenter(samples=samples_in, sample_rate=16000)
    rms_after = np.sqrt(np.mean(samples_out**2))

    assert samples_out.shape == samples_in.shape
    assert samples_out.dtype == np.float32
    assert not (float(rms_after) == pytest.approx(0.0))
    assert rms_after != rms_before


@pytest.mark.parametrize("color", NOISE_COLOR_DECAYS.keys())
def test_noise_has_zero_mean_unit_std(color):
    np.random.seed(42)
    samples_in = np.random.normal(0, 1, size=16000).astype(np.float32)
    f_decay = NOISE_COLOR_DECAYS[color]
    augmenter = AddColorNoise(
        min_f_decay=f_decay,
        max_f_decay=f_decay,
        min_snr_db=0.0,
        max_snr_db=0.0,
        p=1.0,
    )
    noise = augmenter(samples=samples_in, sample_rate=16000) - samples_in
    assert noise.dtype == np.float32
    assert np.mean(noise) == pytest.approx(0.0, abs=0.1)
    assert np.std(noise) == pytest.approx(1.0, abs=0.1)


@pytest.mark.parametrize("color", NOISE_COLOR_DECAYS.keys())
def test_add_colored_noise_snr(color):
    np.random.seed(42)
    samples_in = np.random.normal(0, 1, size=1024).astype(np.float32)

    augmenter = AddColorNoise(min_snr_db=5, max_snr_db=5, p=1.0)
    rms_in = np.sqrt(np.mean(samples_in**2))
    samples_out = augmenter(samples=samples_in, sample_rate=16000)
    rms_noise = np.sqrt(np.mean((samples_out - samples_in) ** 2))
    snr_db = 20 * np.log10(rms_in / rms_noise)

    assert samples_out.dtype == np.float32
    assert snr_db == pytest.approx(5.0, rel=1e-2)


@pytest.mark.parametrize("color", NOISE_COLOR_DECAYS.keys())
def test_add_colored_noise_colors(color):
    np.random.seed(42)
    samples_in = np.random.normal(0, 1, size=16000).astype(np.float32)
    f_decay = NOISE_COLOR_DECAYS[color]
    augmenter = AddColorNoise(min_f_decay=f_decay, max_f_decay=f_decay, p=1.0)
    samples_out = augmenter(samples=samples_in, sample_rate=16000)
    noise = samples_out - samples_in

    # Compute decay rate
    decay_rate = calculate_decay_rate(noise)

    # TODO: Decrease the value of abs= below. To do that, we need a better
    #       calculate_decay_rate function.
    assert decay_rate == pytest.approx(f_decay, abs=1.6)


@pytest.mark.parametrize("color", NOISE_COLOR_DECAYS.keys())
@pytest.mark.parametrize("a_weighted", [False, True])
@pytest.mark.parametrize("size", [(16_000), (2, 16_000)])
def test_generate_decaying_white_noise(color, a_weighted, size):
    np.random.seed(42)
    f_decay = NOISE_COLOR_DECAYS[color]

    noise = generate_decaying_white_noise(
        size=size,
        beta=f_decay,
        apply_a_weighting=a_weighted,
        n_fft=64,
        sample_rate=16000,
    )

    assert np.mean(noise) == pytest.approx(0.0, abs=0.1)
    assert calculate_rms(noise) == pytest.approx(1.0, abs=0.1)

    # TODO: is abs=0.313 below good enough?
    assert noise_kl_divergence_from_white_noise(noise.flatten()) == pytest.approx(
        0.0, abs=0.313
    )



from __future__ import annotations

import warnings
from abc import ABC, ABCMeta, abstractmethod
from typing import Any

SERIALIZABLE_REGISTRY: dict[str, SerializableMeta] = {}
NON_SERIALIZABLE_REGISTRY: dict[str, SerializableMeta] = {}


def shorten_class_name(class_fullname: str) -> str:
    split = class_fullname.split(".")
    if len(split) == 1:
        return class_fullname
    top_module, *_, class_name = split
    if top_module == "audiomentations":
        return class_name
    return class_fullname
    
class SerializableMeta(ABCMeta):
    """A metaclass that is used to register classes in `SERIALIZABLE_REGISTRY` or `NON_SERIALIZABLE_REGISTRY`
    so they can be found later while deserializing transformation pipeline using classes full names.
    """

    def __new__(cls, name: str, bases: tuple[type, ...], *args: Any, **kwargs: Any) -> SerializableMeta:
        cls_obj = super().__new__(cls, name, bases, *args, **kwargs)
        if name != "Serializable" and ABC not in bases:
            if cls_obj.is_serializable():
                SERIALIZABLE_REGISTRY[cls_obj.get_class_fullname()] = cls_obj
            else:
                NON_SERIALIZABLE_REGISTRY[cls_obj.get_class_fullname()] = cls_obj
        return cls_obj

    @classmethod
    def is_serializable(cls) -> bool:
        return False

    @classmethod
    def get_class_fullname(cls) -> str:
        return get_shortest_class_fullname(cls)

    @classmethod
    def _to_dict(cls) -> dict[str, Any]:
        return {}


class Serializable(metaclass=SerializableMeta):
    @classmethod
    @abstractmethod
    def is_serializable(cls) -> bool:
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def get_class_fullname(cls) -> str:
        raise NotImplementedError

    @abstractmethod
    def to_dict_private(self) -> dict[str, Any]:
        raise NotImplementedError

    def to_dict(self, on_not_implemented_error: str = "raise") -> dict[str, Any]:
        """Take a transform pipeline and convert it to a serializable representation that uses only standard
        python data types: dictionaries, lists, strings, integers, and floats.

        Args:
            self: A transform that should be serialized. If the transform doesn't implement the `to_dict`
                method and `on_not_implemented_error` equals to 'raise' then `NotImplementedError` is raised.
                If `on_not_implemented_error` equals to 'warn' then `NotImplementedError` will be ignored
                but no transform parameters will be serialized.
            on_not_implemented_error (str): `raise` or `warn`.

        """
        if on_not_implemented_error not in {"raise", "warn"}:
            msg = f"Unknown on_not_implemented_error value: {on_not_implemented_error}. Supported values are: 'raise' "
            "and 'warn'"
            raise ValueError(msg)
        try:
            transform_dict = self.to_dict_private()
        except NotImplementedError:
            if on_not_implemented_error == "raise":
                raise

            transform_dict = {}
            warnings.warn(
                f"Got NotImplementedError while trying to serialize {self}. Object arguments are not preserved. ",
                stacklevel=2,
            )
        return {"transform": transform_dict}

def get_shortest_class_fullname(cls: type[Any]) -> str:
    """The function `get_shortest_class_fullname` takes a class object as input and returns its shortened
    full name.

    :param cls: The parameter `cls` is of type `Type[BasicCompose]`, which means it expects a class that
    is a subclass of `BasicCompose`
    :type cls: Type[BasicCompose]
    :return: a string, which is the shortened version of the full class name.
    """
    class_fullname = f"{cls.__module__}.{cls.__name__}"
    return shorten_class_name(class_fullname)



import os
import pytest

import numpy as np
from numpy.testing import assert_array_equal, assert_array_almost_equal, assert_allclose

from audiomentations import (
    ClippingDistortion,
    AddBackgroundNoise,
    TimeMask,
    Shift,
    PolarityInversion,
    Gain,
    OneOf,
)
from demo.demo import DEMO_DIR


def test_exactly_one_transform_applied():
    samples = np.array([0.25, 0.0, 0.1, -0.4], dtype=np.float32)
    sample_rate = 44100

    for _ in range(30):
        augmenter = OneOf(
            [
                Gain(min_gain_db=-12, max_gain_db=-6, p=1.0),
                PolarityInversion(p=1.0),
            ]
        )
        perturbed_samples = augmenter(samples=samples, sample_rate=sample_rate)

        was_gain_applied = 0.0 < abs(perturbed_samples[0]) < 0.25
        was_polarity_inversion_applied = perturbed_samples[0] < 0.0

        num_transforms_applied = sum(
            [
                1 if was_gain_applied else 0,
                1 if was_polarity_inversion_applied else 0,
            ]
        )
        assert num_transforms_applied == 1

def test_freeze_and_unfreeze_all_parameters():
    samples = np.array([0.25, 0.0, 0.1, -0.4], dtype=np.float32)
    sample_rate = 44100

    for _ in range(30):
        augmenter = OneOf(
            [
                Gain(p=1.0),
                PolarityInversion(p=1.0),
            ]
        )
        perturbed_samples1 = augmenter(samples=samples, sample_rate=sample_rate)
        augmenter.freeze_parameters()
        for transform in augmenter.transforms:
            assert transform.are_parameters_frozen
        perturbed_samples2 = augmenter(samples=samples, sample_rate=sample_rate)

        assert_array_equal(perturbed_samples1, perturbed_samples2)

        augmenter.unfreeze_parameters()
        for transform in augmenter.transforms:
            assert not transform.are_parameters_frozen

def test_freeze_and_unfreeze_own_parameters():
    augmenter = OneOf(
        [
            Gain(p=1.0),
            PolarityInversion(p=1.0),
        ]
    )
    assert not augmenter.are_parameters_frozen
    for transform in augmenter.transforms:
        assert not transform.are_parameters_frozen

    augmenter.freeze_parameters(apply_to_children=False)
    assert augmenter.are_parameters_frozen
    for transform in augmenter.transforms:
        assert not transform.are_parameters_frozen

    augmenter.unfreeze_parameters(apply_to_children=False)
    assert not augmenter.are_parameters_frozen
    for transform in augmenter.transforms:
        assert not transform.are_parameters_frozen

def test_randomize_all_parameters_and_apply():
    samples = 1.0 / np.arange(1, 21, dtype=np.float32)
    sample_rate = 44100

    augmenter = OneOf(
        [
            AddBackgroundNoise(
                sounds_path=os.path.join(DEMO_DIR, "background_noises"),
                min_snr_db=15,
                max_snr_db=35,
                p=1.0,
            ),
            ClippingDistortion(p=1.0),
            TimeMask(min_band_part=0.2, max_band_part=0.5, p=1.0),
            Shift(min_shift=0.5, max_shift=0.5, p=1.0),
        ]
    )
    augmenter.freeze_parameters()
    augmenter.randomize_parameters(samples=samples, sample_rate=sample_rate)

    parameters = [transform.parameters for transform in augmenter.transforms]

    perturbed_samples1 = augmenter(samples=samples, sample_rate=sample_rate)
    perturbed_samples2 = augmenter(samples=samples, sample_rate=sample_rate)

    assert_array_equal(perturbed_samples1, perturbed_samples2)

    augmenter.unfreeze_parameters()

    for transform_parameters, transform in zip(parameters, augmenter.transforms):
        assert transform_parameters == transform.parameters
        assert not transform.are_parameters_frozen

def test_randomize_only_own_parameters():
    samples = 1.0 / np.arange(1, 21, dtype=np.float32)
    sample_rate = 44100

    augmenter = OneOf(
        [
            Gain(p=1.0),
            PolarityInversion(p=1.0),
        ]
    )
    augmenter.randomize_parameters(samples, sample_rate, apply_to_children=True)
    own_parameters_before = (augmenter.should_apply, augmenter.transform_index)

    parameters = [transform.parameters for transform in augmenter.transforms]

    own_parameters_changed = False
    for i in range(30):
        augmenter.randomize_parameters(
            samples, sample_rate, apply_to_children=False
        )
        if (
            augmenter.should_apply,
            augmenter.transform_index,
        ) != own_parameters_before:
            own_parameters_changed = True
            break
    assert own_parameters_changed

    # Check that the children's parameters are still the same
    for transform_parameters, transform in zip(parameters, augmenter.transforms):
        assert transform_parameters == transform.parameters
        assert not transform.are_parameters_frozen

def test_one_of_weights():
    samples = np.array([1.0, 0.5, -0.25, -0.125, 0.0], dtype=np.float32)
    sample_rate = 16000
    
    transforms = [
        Gain(min_gain_db=6.0, max_gain_db=6.0, p=1.0),  # Gain approx 2.0
        Gain(min_gain_db=-6.0, max_gain_db=-6.0, p=1.0), # Gain approx 0.5
    ]
    weights = [0.8, 0.2]
    
    augmenter = OneOf(transforms=transforms, p=1.0, weights=weights)
    
    counts = [0, 0]
    num_runs = 2000
    for _ in range(num_runs):
        perturbed_samples = augmenter(samples=samples, sample_rate=sample_rate)
        # Check which transform was applied based on the gain effect
        if np.allclose(perturbed_samples[0], samples[0] * 1.99526, atol=1e-4):
            counts[0] += 1
        elif np.allclose(perturbed_samples[0], samples[0] * 0.501187, atol=1e-4):
            counts[1] += 1
        else:
             # This should not happen if p=1.0 and gains are fixed
             raise AssertionError("Unexpected output sample value")

    observed_proportions = np.array(counts) / num_runs
    assert_allclose(observed_proportions, weights, atol=0.05)


def test_one_of_weights_normalization():
    samples = np.array([1.0], dtype=np.float32)
    sample_rate = 16000
    
    transforms = [
        Gain(min_gain_db=6.0, max_gain_db=6.0, p=1.0), 
        Gain(min_gain_db=-6.0, max_gain_db=-6.0, p=1.0),
    ]
    # Weights don't sum to 1, should be normalized to [0.8, 0.2]
    weights = [4, 1] 
    expected_normalized_weights = [0.8, 0.2]

    augmenter = OneOf(transforms=transforms, p=1.0, weights=weights)
    
    counts = [0, 0]
    num_runs = 2000
    for _ in range(num_runs):
        perturbed_samples = augmenter(samples=samples, sample_rate=sample_rate)
        if np.allclose(perturbed_samples[0], samples[0] * 1.99526, atol=1e-4):
            counts[0] += 1
        elif np.allclose(perturbed_samples[0], samples[0] * 0.501187, atol=1e-4):
            counts[1] += 1
        else:
            # This should not happen if p=1.0 and gains are fixed
            raise AssertionError("Unexpected output sample value")

    observed_proportions = np.array(counts) / num_runs
    # Check proportions match the *normalized* weights
    assert_allclose(observed_proportions, expected_normalized_weights, atol=0.05)


def test_one_of_invalid_weights():
    transforms = [
        Gain(p=1.0),
        PolarityInversion(p=1.0),
    ]

    # Length mismatch
    with pytest.raises(ValueError, match="Length of weights must match"):
        OneOf(transforms=transforms, weights=[0.5])
        
    # Negative weights
    with pytest.raises(ValueError, match="weights must be non-negative"):
        OneOf(transforms=transforms, weights=[0.5, -0.1])

    # Weights sum to zero
    with pytest.raises(ValueError, match="Sum of weights must be > 0"):
        OneOf(transforms=transforms, weights=[0.0, 0.0])



from audiomentations.augmentations.base_butterword_filter import BaseButterworthFilter


class BandStopFilter(BaseButterworthFilter):
    """
    Apply band-stop filtering to the input audio. Also known as notch filter or
    band reject filter. It relates to the frequency mask idea in the SpecAugment paper.
    Center frequency gets picked in mel space, so it is
    more aligned with human hearing, which is not linear. Filter steepness
    (6/12/18... dB / octave) is parametrized. Can also be set for zero-phase filtering
    (will result in a 6 dB drop at cutoffs).
    """

    supports_multichannel = True

    def __init__(
        self,
        min_center_freq: float = 200.0,
        max_center_freq: float = 4000.0,
        min_bandwidth_fraction: float = 0.5,
        max_bandwidth_fraction: float = 1.99,
        min_rolloff: int = 12,
        max_rolloff: int = 24,
        zero_phase: bool = False,
        p: float = 0.5,
    ):
        """
        :param min_center_freq: Minimum center frequency in hertz
        :param max_center_freq: Maximum center frequency in hertz
        :param min_bandwidth_fraction: Minimum bandwidth fraction relative to center
            frequency (number between 0 and 2)
        :param max_bandwidth_fraction: Maximum bandwidth fraction relative to center
            frequency (number between 0 and 2)
        :param min_rolloff: Minimum filter roll-off (in dB/octave).
            Must be a multiple of 6
        :param max_rolloff: Maximum filter roll-off (in dB/octave)
            Must be a multiple of 6
        :param zero_phase: Whether filtering should be zero phase.
            When this is set to `True`, it will not affect the phase of the
            input signal but will sound 3 dB lower at the cutoff frequency
            compared to the non-zero phase case (6 dB vs 3 dB). Additionally,
            it is twice as slow as the non-zero phase case. If you
            absolutely want no phase distortions (e.g. want to augment a
            drum track), set this to `True`.
        :param p: The probability of applying this transform
        """
        super().__init__(
            min_center_freq=min_center_freq,
            max_center_freq=max_center_freq,
            min_bandwidth_fraction=min_bandwidth_fraction,
            max_bandwidth_fraction=max_bandwidth_fraction,
            min_rolloff=min_rolloff,
            max_rolloff=max_rolloff,
            zero_phase=zero_phase,
            p=p,
            filter_type="bandstop",
        )



import random

import numpy as np
from numpy.typing import NDArray

from audiomentations.core.transforms_interface import BaseWaveformTransform
from audiomentations.core.utils import calculate_rms


class TanhDistortion(BaseWaveformTransform):
    """
    Apply tanh (hyperbolic tangent) distortion to the audio. This technique is sometimes
    used for adding distortion to guitar recordings. The tanh() function can give a rounded
    "soft clipping" kind of distortion, and the distortion amount is proportional to the
    loudness of the input and the pre-gain. Tanh is symmetric, so the positive and
    negative parts of the signal are squashed in the same way. This transform can be
    useful as data augmentation because it adds harmonics. In other words, it changes
    the timbre of the sound.

    See this page for examples: http://gdsp.hf.ntnu.no/lessons/3/17/
    """

    supports_multichannel = True

    def __init__(
        self, min_distortion: float = 0.01, max_distortion: float = 0.7, p: float = 0.5
    ):
        """
        :param min_distortion: Minimum amount of distortion (between 0 and 1)
        :param max_distortion: Maximum amount of distortion (between 0 and 1)
        :param p: The probability of applying this transform
        """
        super().__init__(p)
        assert 0 <= min_distortion <= 1
        assert 0 <= max_distortion <= 1
        assert min_distortion <= max_distortion
        self.min_distortion = min_distortion
        self.max_distortion = max_distortion

    def randomize_parameters(self, samples: NDArray[np.float32], sample_rate: int):
        super().randomize_parameters(samples, sample_rate)
        if self.parameters["should_apply"]:
            self.parameters["distortion_amount"] = random.uniform(
                self.min_distortion, self.max_distortion
            )

    def apply(self, samples: NDArray[np.float32], sample_rate: int) -> NDArray[np.float32]:
        # Find out how much to pre-gain the audio to get a given amount of distortion
        percentile = 100 - 99 * self.parameters["distortion_amount"]
        threshold = np.percentile(np.abs(samples), percentile)
        gain_factor = 0.5 / (threshold + 1e-6)

        # Distort the audio
        distorted_samples = np.tanh(gain_factor * samples)

        # Scale the output so its loudness matches the input
        rms_before = calculate_rms(samples)
        if rms_before > 1e-9:
            rms_after = calculate_rms(distorted_samples)
            post_gain = rms_before / rms_after
            distorted_samples = post_gain * distorted_samples

        return distorted_samples



import pytest
import numpy as np
from numpy.testing import assert_allclose, assert_array_equal, assert_equal

from audiomentations.core.sampling import WeightedChoiceSampler


class TestWeightedChoiceSampler:
    def test_init_valid_weights(self):
        weights = [0.2, 0.8]
        sampler = WeightedChoiceSampler(weights=weights)
        assert sampler.num_items == 2
        assert_allclose(sampler.cdf, [0.2, 1.0])
        assert sampler.cdf[-1] == 1.0  # Ensure last element is exactly 1.0

    def test_init_valid_weights_needs_normalization(self):
        weights = [1, 4] # Sums to 5
        sampler = WeightedChoiceSampler(weights=weights)
        assert sampler.num_items == 2
        assert_allclose(sampler.cdf, [0.2, 1.0])
        assert sampler.cdf[-1] == 1.0

    def test_init_invalid_weights_none(self):
        with pytest.raises(ValueError, match="If 'weights' is None you must supply a positive 'num_items'."):
            WeightedChoiceSampler(weights=None)

    def test_init_invalid_weights_empty(self):
        with pytest.raises(ValueError, match="Sum of weights must be > 0"):
            WeightedChoiceSampler(weights=[])

    def test_init_invalid_weights_negative(self):
        with pytest.raises(ValueError, match="weights must be non-negative"):
            WeightedChoiceSampler(weights=[0.5, -0.1])

    def test_init_weights_all_zero(self):
        with pytest.raises(ValueError, match="Sum of weights must be > 0"):
            WeightedChoiceSampler(weights=[0.0, 0.0])

    def test_sample_single(self):
        weights = [0.1, 0.2, 0.7]
        sampler = WeightedChoiceSampler(weights=weights)
        sample_index = sampler.sample(size=1)[0]
        assert 0 <= sample_index < 3

    def test_sample_multiple(self):
        weights = [0.1, 0.2, 0.7]
        sampler = WeightedChoiceSampler(weights=weights)
        sample_indices = sampler.sample(size=100)
        assert len(sample_indices) == 100
        assert all(0 <= idx < 3 for idx in sample_indices)

    def test_sample_distribution(self):
        weights = [0.1, 0.3, 0.6]
        sampler = WeightedChoiceSampler(weights=weights)
        num_samples = 20000
        sample_indices = sampler.sample(size=num_samples)

        counts = np.bincount(sample_indices, minlength=len(weights))
        observed_proportions = counts / num_samples

        normalized_weights = np.array(weights) / np.sum(weights)

        # Use a tolerance suitable for statistical tests
        assert_allclose(observed_proportions, normalized_weights, atol=0.02)

    def test_sample_edge_case_one_hot_first(self):
        weights = [1.0, 0.0, 0.0]
        sampler = WeightedChoiceSampler(weights=weights)
        num_samples = 100
        sample_indices = sampler.sample(size=num_samples)
        assert all(idx == 0 for idx in sample_indices)

    def test_sample_edge_case_one_hot_middle(self):
        weights = [0.0, 1.0, 0.0]
        sampler = WeightedChoiceSampler(weights=weights)
        num_samples = 100
        sample_indices = sampler.sample(size=num_samples)
        assert all(idx == 1 for idx in sample_indices)

    def test_sample_edge_case_uniform(self):
        weights = [1.0, 1.0, 1.0]
        sampler = WeightedChoiceSampler(weights=weights)
        num_samples = 30000
        sample_indices = sampler.sample(size=num_samples)

        counts = np.bincount(sample_indices, minlength=len(weights))
        observed_proportions = counts / num_samples

        expected_proportions = np.array([1/3, 1/3, 1/3])

        assert_allclose(observed_proportions, expected_proportions, atol=0.02)

    def test_sample_distribution_uniform(self):
        # Use normalized weights for testing distribution
        weights = np.array([0.5, 0.5])
        sampler = WeightedChoiceSampler(weights=weights)
        num_samples = 10000
        sample_indices = sampler.sample(size=num_samples)
        
        counts = np.bincount(sample_indices, minlength=sampler.num_items)
        observed_proportions = counts / num_samples
        
        assert_allclose(observed_proportions, weights, atol=0.05)

    def test_sample_distribution_non_uniform(self):
         # Use normalized weights for testing distribution
        weights = np.array([0.1, 0.7, 0.2])
        sampler = WeightedChoiceSampler(weights=weights)
        num_samples = 20000
        sample_indices = sampler.sample(size=num_samples)

        counts = np.bincount(sample_indices, minlength=sampler.num_items)
        observed_proportions = counts / num_samples
        
        assert_allclose(observed_proportions, weights, atol=0.05)

    def test_sample_distribution_with_zero_weight(self):
         # Use normalized weights for testing distribution
        weights = np.array([0.5, 0.0, 0.5])
        sampler = WeightedChoiceSampler(weights=weights)
        num_samples = 10000
        sample_indices = sampler.sample(size=num_samples)

        counts = np.bincount(sample_indices, minlength=sampler.num_items)
        observed_proportions = counts / num_samples
        
        assert counts[1] == 0 # Index 1 should never be chosen
        assert_allclose(observed_proportions, weights, atol=0.05)

    def test_sample_single_weight(self):
        weights = [1.0]
        sampler = WeightedChoiceSampler(weights=weights)
        num_samples = 100
        sample_indices = sampler.sample(size=num_samples)
        assert_array_equal(sample_indices, np.zeros(num_samples, dtype=int))
        assert sampler.cdf[-1] == 1.0 



import numpy as np
import pytest

from audiomentations import BitCrush


def test_single_channel():
    samples = np.random.normal(0, 0.1, size=(2048,)).astype(np.float32)
    sample_rate = 16000
    augmenter = BitCrush(min_bit_depth=3, max_bit_depth=6, p=1.0)

    distorted_samples = augmenter(samples=samples, sample_rate=sample_rate)

    assert samples.dtype == distorted_samples.dtype
    assert samples.shape == distorted_samples.shape
    assert 2 ** augmenter.parameters["bit_depth"] + 1 > len(
        np.unique(np.round(distorted_samples, 5))
    )


def test_multichannel():
    num_channels = 3
    samples = np.random.normal(0, 0.1, size=(num_channels, 2048)).astype(np.float32)
    sample_rate = 16000
    augmenter = BitCrush(min_bit_depth=3, max_bit_depth=6, p=1.0)

    distorted_samples = augmenter(samples=samples, sample_rate=sample_rate)

    assert samples.dtype == distorted_samples.dtype
    assert samples.shape == distorted_samples.shape
    assert 2 ** augmenter.parameters["bit_depth"] + 1 > len(
        np.unique(np.round(distorted_samples, 5))
    )


def test_param_range():
    with pytest.raises(ValueError):
        BitCrush(min_bit_depth=0, max_bit_depth=6, p=1.0)
    with pytest.raises(ValueError):
        BitCrush(min_bit_depth=7, max_bit_depth=64, p=1.0)
    with pytest.raises(ValueError):
        BitCrush(min_bit_depth=8, max_bit_depth=6, p=1.0)



import warnings

from audiomentations.augmentations.air_absorption import get_temperature_humidity_key

DEBUG = False

from audiomentations import AirAbsorption
import numpy as np
import pytest
import scipy
import scipy.signal


def get_chirp_test(sample_rate, duration):
    """Create a `duration` seconds chirp from 0Hz to `Nyquist frequency`"""
    n = np.arange(0, duration, 1 / sample_rate)
    samples = scipy.signal.chirp(n, 0, duration, sample_rate // 2, method="linear")
    return samples.astype(np.float32)


@pytest.mark.parametrize("temperature", [10, 20])
@pytest.mark.parametrize("humidity", [35, 55, 75])
@pytest.mark.parametrize("distance", [5, 10, 20, 100, 2500])
@pytest.mark.parametrize("sample_rate", [8000, 16000, 48000])
def test_input_shapes(temperature, humidity, distance, sample_rate):
    np.random.seed(1)

    samples = get_chirp_test(sample_rate, 10)

    augment = AirAbsorption(
        min_temperature=temperature,
        max_temperature=temperature,
        min_humidity=humidity,
        max_humidity=humidity,
        min_distance=distance,
        max_distance=distance,
        p=1.0,
    )

    with warnings.catch_warnings():
        # Assert that no warning is emitted
        warnings.simplefilter("error")

        # Test 1D case
        processed_samples = augment(samples, sample_rate=sample_rate)
        assert processed_samples.shape == samples.shape
        assert processed_samples.dtype == np.float32

        # Test 2D case
        samples = np.tile(samples, (2, 1))
        processed_samples = augment(samples, sample_rate=sample_rate)
        assert processed_samples.shape == samples.shape
        assert processed_samples.dtype == np.float32


def test_get_temperature_humidity_key():
    assert get_temperature_humidity_key(10.0, 30.0) == "10C_30-50%"
    assert get_temperature_humidity_key(10.0, 45.0) == "10C_30-50%"
    assert get_temperature_humidity_key(10.0, 50.0) == "10C_30-50%"
    assert get_temperature_humidity_key(10.0, 51.0) == "10C_50-70%"
    assert get_temperature_humidity_key(10.0, 65.0) == "10C_50-70%"
    assert get_temperature_humidity_key(20.0, 90.0) == "20C_70-90%"


def test_validate_parameters():
    with pytest.raises(ValueError):
        AirAbsorption(min_temperature=0.0)
    with pytest.raises(ValueError):
        AirAbsorption(max_temperature=-90.0)
    with pytest.raises(ValueError):
        AirAbsorption(min_temperature=20.0, max_temperature=10.0)
    with pytest.raises(ValueError):
        AirAbsorption(min_humidity=10.0)
    with pytest.raises(ValueError):
        AirAbsorption(min_distance=0.0)
    with pytest.raises(ValueError):
        AirAbsorption(min_distance=4444.0)
    with pytest.raises(ValueError):
        AirAbsorption(max_distance=4000)
    with pytest.raises(ValueError):
        AirAbsorption(max_distance=-50)
    with pytest.raises(ValueError):
        AirAbsorption(min_distance=40.0, max_distance=10.0)



import numpy as np
import pytest

from audiomentations import ClippingDistortion


def test_distort():
    sample_len = 1024
    samples_in = np.random.normal(0, 1, size=sample_len).astype(np.float32)
    sample_rate = 16000
    augmenter = ClippingDistortion(
        min_percentile_threshold=20, max_percentile_threshold=40, p=1.0
    )

    samples_out = augmenter(samples=samples_in, sample_rate=sample_rate)
    assert samples_out.dtype == np.float32
    assert len(samples_out) == sample_len
    assert sum(abs(samples_out)) < sum(abs(samples_in))


def test_distort_multichannel():
    sample_len = 32000
    samples_in = np.random.normal(0, 1, size=(2, sample_len)).astype(np.float32)
    sample_rate = 16000
    augmenter = ClippingDistortion(
        min_percentile_threshold=20, max_percentile_threshold=40, p=1.0
    )

    samples_out = augmenter(samples=samples_in, sample_rate=sample_rate)
    assert samples_out.dtype == np.float32
    assert samples_out.shape == samples_in.shape
    assert np.sum(np.abs(samples_out)) < np.sum(np.abs(samples_in))
    assert np.amax(samples_out[0, :]) == pytest.approx(np.amax(samples_out[1, :]))



import pytest
import numpy as np
from numpy.testing import assert_almost_equal

from audiomentations import LoudnessNormalization


def test_loudness_normalization():
    samples = np.random.uniform(low=-0.2, high=-0.001, size=(8000,)).astype(np.float32)
    sample_rate = 16000

    augment = LoudnessNormalization(min_lufs=-32, max_lufs=-12, p=1.0)
    processed_samples = augment(samples=samples, sample_rate=sample_rate)
    gain_factors = processed_samples / samples
    assert np.amin(gain_factors) == pytest.approx(np.amax(gain_factors))
    assert processed_samples.dtype == np.float32


def test_loudness_normalization_digital_silence():
    samples = np.zeros(8000, dtype=np.float32)
    sample_rate = 16000

    augment = LoudnessNormalization(min_lufs=-32, max_lufs=-12, p=1.0)
    processed_samples = augment(samples=samples, sample_rate=sample_rate)
    assert_almost_equal(processed_samples, np.zeros(8000, dtype=np.float32))
    assert processed_samples.dtype == np.float32


def test_loudness_normalization_too_short_input():
    samples = np.random.uniform(low=-0.2, high=-0.001, size=(800,)).astype(np.float32)
    sample_rate = 16000

    augment = LoudnessNormalization(min_lufs=-32, max_lufs=-12, p=1.0)
    with pytest.raises(ValueError):
        _ = augment(samples=samples, sample_rate=sample_rate)


def test_loudness_normalization_multichannel():
    samples = np.random.uniform(low=-0.2, high=-0.001, size=(3, 8000)).astype(
        np.float32
    )
    sample_rate = 16000

    augment = LoudnessNormalization(min_lufs=-32, max_lufs=-12, p=1.0)
    processed_samples = augment(samples=samples, sample_rate=sample_rate)
    gain_factors = processed_samples / samples
    assert np.amin(gain_factors) == pytest.approx(np.amax(gain_factors))
    assert processed_samples.dtype == np.float32


def test_validation():
    with pytest.raises(ValueError):
        LoudnessNormalization(min_lufs=-12.0, max_lufs=-32.0)



import numpy as np
import pytest

from audiomentations import TimeMask


def test_apply_time_mask():
    sample_len = 1024
    samples_in = np.random.normal(0, 1, size=sample_len).astype(np.float32)
    sample_rate = 16000
    augmenter = TimeMask(min_band_part=0.2, max_band_part=0.5, fade_duration=0.0, p=1.0)

    samples_out = augmenter(samples=samples_in, sample_rate=sample_rate)
    assert samples_out.dtype == np.float32
    assert len(samples_out) == sample_len

    std_in = np.mean(np.abs(samples_in))
    std_out = np.mean(np.abs(samples_out))
    assert std_out < std_in


def test_apply_time_mask_start():
    sample_len = 1000
    samples_in = np.random.normal(0, 1, size=sample_len).astype(np.float32)
    sample_rate = 8000
    augmenter = TimeMask(min_band_part=0.1, max_band_part=0.1, fade_duration=0.0, mask_location="start", p=1.0)

    samples_out = augmenter(samples=samples_in, sample_rate=sample_rate)
    assert samples_out.dtype == np.float32
    assert len(samples_out) == sample_len
    assert not np.any(samples_out[0:100])
    assert samples_out[101] == samples_in[101]

def test_apply_time_mask_end():
    sample_len = 1000
    samples_in = np.random.normal(0, 1, size=sample_len).astype(np.float32)
    sample_rate = 8000
    augmenter = TimeMask(min_band_part=0.1, max_band_part=0.1, fade_duration=0.0, mask_location="end", p=1.0)

    samples_out = augmenter(samples=samples_in, sample_rate=sample_rate)
    assert samples_out.dtype == np.float32
    assert len(samples_out) == sample_len
    assert not np.any(samples_out[-100:])
    assert samples_out[-101] == samples_in[-101]


def test_invalid_params():
    with pytest.raises(ValueError):
        TimeMask(min_band_part=0.5, max_band_part=1.5)

    with pytest.raises(ValueError):
        TimeMask(min_band_part=-0.5, max_band_part=0.5)

    with pytest.raises(ValueError):
        TimeMask(min_band_part=0.6, max_band_part=0.5)

    with pytest.raises(ValueError):
        TimeMask(fade_duration=-0.1)

    with pytest.raises(ValueError):
        TimeMask(fade_duration=0.00001)

    with pytest.raises(ValueError):
        TimeMask(mask_location="beyond_infinity")

    with pytest.raises(TypeError):
        TimeMask(fade=True)


def test_apply_time_mask_multichannel():
    sample_len = 1024
    samples_in = np.random.normal(0, 1, size=(2, sample_len)).astype(np.float32)
    sample_rate = 16000
    augmenter = TimeMask(min_band_part=0.2, max_band_part=0.5, p=1.0)

    samples_out = augmenter(samples=samples_in, sample_rate=sample_rate)
    assert samples_out.dtype == np.float32
    assert samples_out.shape == samples_in.shape

    std_in = np.mean(np.abs(samples_in))
    std_out = np.mean(np.abs(samples_out))
    assert std_out < std_in


def test_apply_time_mask_with_fade():
    sample_len = 1024
    samples_in = np.random.normal(0, 1, size=sample_len).astype(np.float32)
    sample_rate = 16000
    augmenter = TimeMask(min_band_part=0.2, max_band_part=0.5, p=1.0)

    samples_out = augmenter(samples=samples_in, sample_rate=sample_rate)
    assert samples_out.dtype == np.float32
    assert len(samples_out) == sample_len

    std_in = np.mean(np.abs(samples_in))
    std_out = np.mean(np.abs(samples_out))
    assert std_out < std_in


def test_apply_time_mask_with_fade_short_signal():
    sample_len = 100
    samples_in = np.random.normal(0, 1, size=sample_len).astype(np.float32)
    sample_rate = 16000
    augmenter = TimeMask(min_band_part=0.2, max_band_part=0.5, p=1.0)

    samples_out = augmenter(samples=samples_in, sample_rate=sample_rate)
    assert samples_out.dtype == np.float32
    assert len(samples_out) == sample_len

    std_in = np.mean(np.abs(samples_in))
    std_out = np.mean(np.abs(samples_out))
    assert std_out < std_in



import random

import numpy as np
import pytest
from numpy import array_equal

from audiomentations import Padding


@pytest.mark.parametrize("mode", ["silence", "wrap", "reflect"])
@pytest.mark.parametrize("pad_section", ["start", "end"])
def test_padding_mono_1d(mode, pad_section):
    random.seed(546)
    samples = np.array([0.5, 0.6, -0.2, 1.0], dtype=np.float32)
    sample_rate = 16000
    input_shape = samples.shape
    augmenter = Padding(mode=mode, pad_section=pad_section, p=1.0)
    samples = augmenter(samples=samples, sample_rate=sample_rate)

    assert samples.dtype == np.float32
    assert samples.shape == input_shape


def test_padding_mono_2d():
    samples = np.array(
        [[0.9, 0.5, -0.25, -0.125, 0.0]],
        dtype=np.float32,
    )
    sample_rate = 16000
    input_shape = samples.shape

    augmenter = Padding(p=1.0)
    samples = augmenter(samples=samples, sample_rate=sample_rate)

    assert samples.dtype == np.float32
    assert samples.shape == input_shape


@pytest.mark.parametrize("mode", ["silence", "wrap", "reflect"])
@pytest.mark.parametrize("pad_section", ["start", "end"])
def test_padding_multichannel(mode, pad_section):
    samples = np.array(
        [
            [0.9, 0.5, -0.25, -0.125, 0.0],
            [0.95, 0.5, -0.25, -0.125, 0.0],
            [0.95, 0.5, -0.25, -0.125, 0.0],
        ],
        dtype=np.float32,
    )
    sample_rate = 16000
    input_shape = samples.shape

    augmenter = Padding(mode=mode, pad_section=pad_section, p=1.0)
    samples = augmenter(samples=samples, sample_rate=sample_rate)

    assert samples.dtype == np.float32
    assert samples.shape == input_shape


def test_padding_reflect_start():
    samples = np.array([0.5, 0.6, 0.9, -0.2, 1.0], dtype=np.float32)
    sample_rate = 16000
    augmenter = Padding(
        mode="reflect",
        pad_section="start",
        min_fraction=0.4,
        max_fraction=0.4,
        p=1.0,
    )
    samples = augmenter(samples=samples, sample_rate=sample_rate)
    assert array_equal(samples, np.array([1.0, -0.2, 0.9, -0.2, 1.0], dtype=np.float32))


def test_padding_reflect_end():
    samples = np.array([0.5, 0.6, 0.9, -0.2, 1.0], dtype=np.float32)
    sample_rate = 16000
    augmenter = Padding(
        mode="reflect",
        pad_section="end",
        min_fraction=0.4,
        max_fraction=0.4,
        p=1.0,
    )
    samples = augmenter(samples=samples, sample_rate=sample_rate)
    assert array_equal(samples, np.array([0.5, 0.6, 0.9, 0.6, 0.5], dtype=np.float32))


def test_pad_nothing():
    samples = np.array([0.5, 0.6, -0.2, 0.1], dtype=np.float32)
    sample_rate = 16000
    input_shape = samples.shape
    augmenter = Padding(min_fraction=0.0, max_fraction=0.0, p=1.0)
    samples = augmenter(samples=samples, sample_rate=sample_rate)

    assert array_equal(samples, np.array([0.5, 0.6, -0.2, 0.1], dtype=np.float32))
    assert samples.dtype == np.float32
    assert samples.shape == input_shape


def test_pad_everything():
    samples = np.array([0.5, 0.6, -0.2, 0.7], dtype=np.float32)
    sample_rate = 16000
    input_shape = samples.shape
    augmenter = Padding(min_fraction=1.0, max_fraction=1.0, p=1.0)
    samples = augmenter(samples=samples, sample_rate=sample_rate)

    assert not np.any(samples)
    assert samples.dtype == np.float32
    assert samples.shape == input_shape



import random

import numpy as np
import scipy.signal as sp
from numpy.typing import NDArray

from audiomentations.core.transforms_interface import BaseWaveformTransform
from audiomentations.core.utils import a_weighting_frequency_envelope
from audiomentations.core.utils import calculate_desired_noise_rms, calculate_rms

NOISE_COLOR_DECAYS = {
    "pink": -1.0 * 10 * np.log10(2),
    "brown": -2.0 * 10 * np.log10(2),
    "brownian": -2.0 * 10 * np.log10(2),
    "red": -2.0 * 10 * np.log10(2),
    "blue": 1.0 * 10 * np.log10(2),
    "azure": 1.0 * 10 * np.log10(2),
    "violet": 2.0 * 10 * np.log10(2),
    "white": 0.0 * 10 * np.log10(2),
}


def decay_to_beta(decay: float) -> float:
    """
    Converts a decay given in dB/octave to the beta exponent
    in the PSD logarithmic decay function:

    logarithmic_decay = \sqrt{\frac{1}{f^{\beta}}}

    where logarithmic_decay is the rate of change of the PSD
    over frequency (in log-space).
    """

    return decay / (-10.0) / np.log10(2.0)


def generate_decaying_white_noise(
    size,
    beta,
    sample_rate,
    apply_a_weighting=False,
    n_fft=64,
    in_db_per_octave=True,
):
    """
    Generates a white noise signal decaying linearly by 1/f^beta
    (when in_db_per_octave==False) or changing (decaying or increasing)
    linearly by beta dB/octave (when in_db_per_octave==False).

    The values for beta are given below:

    | Colour         |  in_db_per_octave=False | in_db_per_octave=True |
    |----------------+-------------------------+-----------------------|
    | pink           |                     1.0 |                 -3.01 |
    | brown/brownian |                     2.0 |                 -6.02 |
    | red            |                     2.0 |                 -6.02 |
    | blue           |                    -1.0 |                  3.01 |
    | azure          |                    -1.0 |                  3.01 |
    | violet         |                    -2.0 |                  6.02 |
    | white          |                     0.0 |                   0.0 |

    Note that you can get away with low n_fft (e.g. 128 points) values
    if you are not using a_weighting, but keep it higher otherwise.
    """

    sig = np.random.normal(0, 1, size=size)

    if beta == 0.0 and not apply_a_weighting:
        # No decay, return white noise
        return sig.astype(np.float32)

    # Compute the decay in fft domain (ignore phase)
    f = np.fft.rfftfreq(n_fft, 1.0 / sample_rate)
    f[0] = 1
    decay = np.ones(n_fft // 2 + 1, dtype=complex)

    # If beta is given as decay in db/octave, convert it to a
    # decay exponent.
    if in_db_per_octave:
        beta = decay_to_beta(beta)

    # Decay is in PSD, for magnitude, take sqrt and add random phase
    decay = np.sqrt(1 / f**beta) * np.exp(
        1j * np.random.uniform(0, 2 * np.pi, len(decay))
    )

    # Optionally apply a-weighting
    if apply_a_weighting:
        weighting = a_weighting_frequency_envelope(n_fft, sample_rate)
        decay *= weighting

    # Compute the impulse response of the decay
    decay_ir = np.fft.irfft(decay)

    # Convolve the white noise with the decay impulse response

    # Calculate number of channels
    if len(sig.shape) > 1:
        n_channels = sig.shape[0]
        fsig = np.zeros_like(sig).astype(np.float32)
        for chn_idx in range(n_channels):
            fsig[chn_idx] = sp.oaconvolve(sig[chn_idx], decay_ir, "same")
    else:
        n_channels = 1
        fsig = sp.oaconvolve(sig, decay_ir, "same")

    # Normalize to unit energy
    fsig /= np.sqrt(np.mean(fsig**2))

    return fsig.astype(np.float32)


class AddColorNoise(BaseWaveformTransform):
    """
    Adds noise to the input samples with a decaying frequency spectrum resulting in "color" noise.
    For info, more see the Wikipedia article here: https://en.wikipedia.org/wiki/Colors_of_noise
    """

    supports_multichannel = True

    def __init__(
        self,
        min_snr_db: float = 5.0,
        max_snr_db: float = 40.0,
        min_f_decay: float = -6.0,
        max_f_decay: float = 6.0,
        p_apply_a_weighting: float = 0.0,
        p: float = 0.5,
        n_fft: int = 128,
    ):
        """
        :param min_snr_db: Minimum signal-to-noise ratio in dB. A lower number means more noise.
        :param max_snr_db: Maximum signal-to-noise ratio in dB. A greater number means less noise.
        :param min_f_decay: Minimum frequency decay in dB per octave.
        :param max_f_decay: Maximum frequency decay in dB per octave.
        :param p: The probability of applying this transform
        :param p_apply_a_weighting: The probability of applying A-weighting to the noise. Useful for
                                    generating "grey" noise. See here: https://en.wikipedia.org/wiki/Grey_noise
        :param n_fft: Noise is 'colorized' by applying a decay in the frequency domain. The decay envelope
                      is generated in the fft domain. This parameter controls the number of fft points. For
                      unweighted colored noise, you should use low n_fft (e.g. 128 points) values, but
                      you should probably increase it higher otherwise.

        f_decay is the decay applied to the spectral density of the noise. The power spectral density per
        unit of bandwidth becomes proportional to 1/f^f_decay. For example, a decay of 0 means that the
        spectral density is flat, equivalent to applying white noise. Below are common noise colors and
        their corresponding f_decay values:

        | Colour   | f_decay (dB/octave)   |
        |----------+-----------------------|
        | pink     |                 -3.01 |
        | brown    |                 -6.02 |
        | Brownian |                 -6.02 |
        | red      |                 -6.02 |
        | blue     |                  3.01 |
        | azure    |                  3.01 |
        | violet   |                  6.02 |
        | white    |                   0.0 |
        For information about those values, see here: https://en.wikipedia.org/wiki/Colors_of_noise

        Additionally, p_apply_a_weighting gives the probability the noise to be weighted by a psychoacoustic
        equal loudness curve, which results in grey-noise when f_decay = 0.0.
        """
        super().__init__(p)
        self.min_snr_db = min_snr_db
        self.max_snr_db = max_snr_db
        self.min_f_decay = min_f_decay
        self.max_f_decay = max_f_decay
        self.p_apply_a_weighting = p_apply_a_weighting
        self.n_fft = n_fft

    def randomize_parameters(self, samples: np.ndarray, sample_rate: int):
        super().randomize_parameters(samples, sample_rate)
        if self.parameters["should_apply"]:
            # Pick SNR in Decibel scale
            snr = random.uniform(self.min_snr_db, self.max_snr_db)

            # Pick f_decay
            f_decay = random.uniform(self.min_f_decay, self.max_f_decay)

            # Pick whether to apply A-weighting
            apply_a_weighting = random.random() < self.p_apply_a_weighting

            # Calculate desired noise rms
            clean_rms = calculate_rms(samples)
            desired_noise_rms = calculate_desired_noise_rms(
                clean_rms=clean_rms, snr=snr
            )

            # Set the parameters
            self.parameters["desired_noise_rms"] = float(desired_noise_rms)
            self.parameters["f_decay"] = f_decay
            self.parameters["apply_a_weighting"] = apply_a_weighting

    def apply(
        self, samples: NDArray[np.float32], sample_rate: int
    ) -> NDArray[np.float32]:
        desired_noise_rms = self.parameters["desired_noise_rms"]

        if samples.ndim == 1:
            n_channels = 1
        else:
            n_channels = samples.shape[0]

        noise_with_unit_rms = generate_decaying_white_noise(
            size=samples.shape,
            beta=self.parameters["f_decay"],
            sample_rate=sample_rate,
            apply_a_weighting=self.parameters["apply_a_weighting"],
            n_fft=self.n_fft,
        )

        if n_channels > 1:
            return samples + noise_with_unit_rms * desired_noise_rms
        else:
            return samples + noise_with_unit_rms * desired_noise_rms
